"""
This model follows the MJD analysis for the peak shape. The peak shape is modeled as the sum of a full-energy Gaussian
component and an exponentially modified Gaussian tail to approximate the peak shape distortion due to incomplete charge
collection. See S. I. Alvis et al., Phys. Rev. C 100, 025501 (2019) for details. In addition, a slightly different
analysis window is used, again following MJD.

Also see J.M. Lopez-Castano and Ping-Han Chu. Energy systematic of majorana demonstrator. Technical report, 2022.
Internal Report, Feb. 9, 2022. for an explanation of the peak shape used.
"""

import sys
from math import erfc

import numba as nb
import numpy as np

import freqfit.models.constants as constants
from freqfit.utils import inspectparameters

nb_kwd = {
    "nopython": True,
    "parallel": False,
    "nogil": True,
    "cache": True,
    "fastmath": True,
    "inline": "always",
}

QBB = constants.QBB
N_A = constants.NA
M_A = constants.MDET

# default analysis window and width
# window
#     uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
#     where edges of window are monotonically increasing (this is not checked), in keV.
#     Default is typical analysis window.
WINDOW = np.array(constants.MJD_WINDOW)

WINDOWSIZE = 0.0
for i in range(len(WINDOW)):
    WINDOWSIZE += WINDOW[i][1] - WINDOW[i][0]

SEED = 42  # set the default random seed

limit = np.log(sys.float_info.max) / 10


# yoinked from https://github.com/legend-exp/pygama/blob/main/src/pygama/math/functions/gauss.py#L72
@nb.jit(**nb_kwd)
def nb_gauss_pdf(x: np.ndarray, mu: float, sigma: float) -> np.ndarray:
    r"""
    Normalised Gaussian PDF, w/ args: mu, sigma. The support is :math:`(-\infty, \infty)`

    .. math::
        pdf(x, \mu, \sigma) = \frac{1}{\sqrt{2\pi}}e^{(\frac{x-\mu}{\sigma}^2)/2}

    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    mu
        The centroid of the Gaussian
    sigma
        The standard deviation of the Gaussian
    """

    if sigma == 0:
        invs = np.inf
    else:
        invs = 1.0 / sigma
    z = (x - mu) * invs
    invnorm = invs / np.sqrt(2 * np.pi)
    return np.exp(-0.5 * z**2) * invnorm


# yoinked from https://github.com/legend-exp/pygama/blob/main/src/pygama/math/functions/exgauss.py
@nb.jit(**nb_kwd)
def nb_gauss_tail_exact(
    x: float, mu: float, sigma: float, tau: float, tmp: float
) -> float:
    r"""
    Exact form of a normalized exponentially modified Gaussian PDF.
    It computes the following PDF:


    .. math::
        pdf(x, \tau,\mu,\sigma) = \frac{1}{2|\tau|}e^{\frac{x-\mu}{\tau}+\frac{\sigma^2}{2\tau^2}}\text{erfc}\left(\frac{\tau(\frac{x-\mu}{\sigma})+\sigma}{|\tau|\sqrt{2}}\right)


    Where :math:`tmp = \frac{x-\mu}{\tau}+\frac{\sigma^2}{2\tau^2}` is precomputed in :func:`nb_exgauss_pdf` to save computational time.


    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.


    Parameters
    ----------
    x
        Input data
    mu
        The centroid of the Gaussian
    sigma
        The standard deviation of the Gaussian
    tau
        The characteristic scale of the Gaussian tail
    tmp
        The scaled version of the exponential argument


    See Also
    --------
    :func:`nb_exgauss_pdf`
    """

    abstau = np.absolute(tau)
    if tmp < limit:
        tmp = tmp
    else:
        tmp = limit
    if sigma == 0 or abstau == 0:
        return x * 0
    z = (x - mu) / sigma
    tail_f = (
        (1 / (2 * abstau))
        * np.exp(tmp)
        * erfc((tau * z + sigma) / (np.sqrt(2) * abstau))
    )
    return tail_f


# yoinked from https://github.com/legend-exp/pygama/blob/main/src/pygama/math/functions/exgauss.py
@nb.jit(**nb_kwd)
def nb_gauss_tail_approx(
    x: np.ndarray, mu: float, sigma: float, tau: float
) -> np.ndarray:
    r"""
    Approximate form of a normalized exponentially modified Gaussian PDF
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        Input data
    mu
        The centroid of the Gaussian
    sigma
        The standard deviation of the Gaussian
    tau
        The characteristic scale of the Gaussian tail


    See Also
    --------
    :func:`nb_exgauss_pdf`
    """
    if sigma == 0:
        return x * 0
    elif (sigma + tau * (x - mu) / sigma) == 0:
        return x * 0
    den = 1 / (sigma + tau * (x - mu) / sigma)
    tail_f = sigma * nb_gauss_pdf(x, mu, sigma) * den * (1.0 - tau * tau * den * den)
    return tail_f


# yoinked from https://github.com/legend-exp/pygama/blob/main/src/pygama/math/functions/exgauss.py
@nb.jit(**nb_kwd)
def nb_exgauss_pdf(x: np.ndarray, mu: float, sigma: float, tau: float) -> np.ndarray:
    r"""
    Normalized PDF of an exponentially modified Gaussian distribution. Its range of support is :math:`x\in(-\infty,\infty)`, :math:`\tau\in(-\infty,\infty)`
    Calls either :func:`nb_gauss_tail_exact` or :func:`nb_gauss_tail_approx` depending on which is computationally cheaper


    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.


    Parameters
    ----------
    x
        Input data
    mu
        The centroid of the Gaussian
    sigma
        The standard deviation of the Gaussian
    tau
        The characteristic scale of the Gaussian tail


    See Also
    --------
    :func:`nb_gauss_tail_exact`, :func:`nb_gauss_tail_approx`
    """

    x = np.asarray(x)
    tail_f = np.empty_like(x, dtype=np.float64)

    for i in nb.prange(x.shape[0]):
        if tau == 0:
            tail_f[i] = np.nan
        else:
            tmp = ((x[i] - mu) / tau) + ((sigma**2) / (2 * tau**2))
            if tmp < limit:
                tail_f[i] = nb_gauss_tail_exact(x[i], mu, sigma, tau, tmp)
            else:
                tail_f[i] = nb_gauss_tail_approx(x[i], mu, sigma, tau)
    return tail_f


@nb.jit(**nb_kwd)
def nb_pdf(
    Es: np.array,
    S: float,
    BI: float,
    frac: float,
    delta: float,
    sigma: float,
    tau: float,
    gamma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
    check_window: bool = False,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    frac
        portion of the peak in the tail
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    tau
        scale parameter of the tail in keV
    gamma
        scaling parameter for tau and sigma
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr
    check_window
        whether to check if the passed Es fall inside of the window. Default is False and assumes that the passed Es
        all fall inside the window (for speed)

    Notes
    -----
    This function computes the following:
    mu_S = (eff + effuncscale * effunc) * exp * S
    mu_B = exp * BI * windowsize
    pdf(E) = 1/(mu_S+mu_B) * [mu_S * norm(E_j, QBB + mu, sigma) + mu_B/windowsize]
    """

    # Precompute the signal and background counts
    # mu_S = np.log(2) * (N_A * S) * (eff + effuncscale * effunc) * exp / M_A
    mu_S = S * (eff + effuncscale * effunc) * exp
    mu_B = exp * BI * WINDOWSIZE

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = exp * BI
    S_amp_gauss = mu_S / (np.sqrt(2 * np.pi) * gamma * sigma) * (1 - frac)

    exgaus = mu_S * frac * nb_exgauss_pdf(Es, QBB + delta, gamma * sigma, gamma * tau)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = (1 / (mu_S + mu_B)) * (
            (
                exgaus[i]
                + S_amp_gauss
                * np.exp(-((Es[i] - QBB - delta) ** 2) / (2 * (gamma * sigma) ** 2))
            )
            + B_amp
        )

    if check_window:
        for i in nb.prange(Es.shape[0]):
            inwindow = False
            for j in range(len(WINDOW)):
                if WINDOW[j][0] <= Es[i] <= WINDOW[j][1]:
                    inwindow = True
            if not inwindow:
                y[i] = 0.0

    return y


@nb.jit(**nb_kwd)
def nb_density(
    Es: np.array,
    S: float,
    BI: float,
    frac: float,
    delta: float,
    sigma: float,
    tau: float,
    gamma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
    check_window: bool = False,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    frac
        portion of the peak in the tail
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    tau
        scale parameter of the tail in keV
    gamma
        scaling parameter for tau and sigma
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr
    check_window
        whether to check if the passed Es fall inside of the window. Default is False and assumes that the passed Es
        all fall inside the window (for speed)

    Notes
    -----
    This function computes the following, faster than without a numba wrapper:
    mu_S = (eff + effuncscale * effunc) * exp * S
    mu_B = exp * BI * windowsize
    CDF(E) = mu_S + mu_B
    pdf(E) =[mu_S * norm(E_j, QBB + mu, sigma) + mu_B/windowsize]
    """

    # Precompute the signal and background counts
    # mu_S = np.log(2) * (N_A * S) * (eff + effuncscale * effunc) * exp / M_A
    mu_S = S * (eff + effuncscale * effunc) * exp
    mu_B = exp * BI * WINDOWSIZE

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = exp * BI
    S_amp_gauss = mu_S / (np.sqrt(2 * np.pi) * gamma * sigma) * (1.0 - frac)

    exgaus = mu_S * frac * nb_exgauss_pdf(Es, QBB + delta, gamma * sigma, gamma * tau)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = (
            exgaus[i]
            + S_amp_gauss
            * np.exp(-((Es[i] - QBB - delta) ** 2) / (2 * (gamma * sigma) ** 2))
        ) + B_amp

    if check_window:
        for i in nb.prange(Es.shape[0]):
            inwindow = False
            for j in range(len(WINDOW)):
                if WINDOW[j][0] <= Es[i] <= WINDOW[j][1]:
                    inwindow = True
            if not inwindow:
                y[i] = 0.0

    return mu_S + mu_B, y


@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_rvs(
    n_sig: int,
    n_bkg: int,
    frac: float,
    delta: float,
    sigma: float,
    tau: float,
    gamma: float,
    seed: int = SEED,
) -> np.array:
    """
    Parameters
    ----------
    n_sig
        Number of signal events to pull from
    n_bkg
        Number of background events to pull from
    frac
        portion of the peak in the tail
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    tau
        scale parameter of the tail in keV
    gamma
        scaling parameter for tau and sigma
    seed
        specify a seed, otherwise uses default seed

    Notes
    -----
    This function pulls from a Gaussian for signal events and from a uniform distribution for background events
    in the provided windows, which may be discontinuous.
    """

    np.random.seed(seed)

    # preallocate for draws
    Es = np.append(np.zeros(n_sig), np.zeros(n_bkg))

    # get signal events from Gaussian with tail
    # first, draw to determine whether from Gaussian or exgaus
    which = np.random.uniform(0, 1, n_sig)
    # draw Gaussian smearing
    smear = np.random.normal(0, gamma * sigma, size=n_sig)
    # draw exponential
    exp = np.random.exponential(scale=gamma * tau, size=n_sig)

    # depending on whether the event should fall in the tail, subtract some energy from Qbb before Gaussian smearing
    for i in range(n_sig):
        if which[i] < frac:
            Es[i] = QBB + delta - exp[i] + smear[i]
        else:
            Es[i] = QBB + delta + smear[i]

    # Get background events from a uniform distribution
    bkg = np.random.uniform(0, 1, n_bkg)

    breaks = np.zeros(shape=(len(WINDOW), 2))
    for i in range(len(WINDOW)):
        thiswidth = WINDOW[i][1] - WINDOW[i][0]

        if i > 0:
            breaks[i][0] = breaks[i - 1][1]

        if i < len(WINDOW):
            breaks[i][1] = breaks[i][0] + thiswidth / WINDOWSIZE

        for j in range(len(bkg)):
            if breaks[i][0] <= bkg[j] <= breaks[i][1]:
                Es[n_sig + j] = (bkg[j] - breaks[i][0]) * thiswidth / (
                    breaks[i][1] - breaks[i][0]
                ) + WINDOW[i][0]

    return Es


@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_extendedrvs(
    S: float,
    BI: float,
    frac: float,
    delta: float,
    sigma: float,
    tau: float,
    gamma: float,
    eff: float,
    effunc: float,
    effuncscale: float,
    exp: float,
    seed: int = SEED,
) -> np.array:
    """
    Parameters
    ----------
    S
        expected rate of signal events in events/(kg*yr)
    BI
        rate of background events in events/(kev*kg*yr)
    frac
        portion of the peak in the tail
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    tau
        scale parameter of the tail in keV
    gamma
        scaling parameter for tau and sigma
    eff
        The global signal efficiency, unitless
    effunc
        uncertainty on the efficiency
    effuncscale
        scaling parameter of the efficiency
    exp
        The exposure, in kg*yr
    seed
        specify a seed, otherwise uses default seed

    Notes
    -----
    This function pulls from a Gaussian for signal events and from a uniform distribution for background events
    in the provided windows, which may be discontinuous.
    """

    n_sig = np.random.poisson(S * (eff + effuncscale * effunc) * exp)
    n_bkg = np.random.poisson(BI * exp * WINDOWSIZE)

    return nb_rvs(n_sig, n_bkg, frac, delta, sigma, tau, gamma, seed=seed), (n_sig, n_bkg)


class mjd_0vbb_gen:
    def __init__(self):
        self.parameters = inspectparameters(self.density)
        del self.parameters["check_window"]  # this should not be seen by iminuit

    def pdf(
        self,
        Es: np.array,
        S: float,
        BI: float,
        frac: float,
        delta: float,
        sigma: float,
        tau: float,
        gamma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        check_window: bool = False,
    ) -> np.array:
        return nb_pdf(
            Es,
            S,
            BI,
            frac,
            delta,
            sigma,
            tau,
            gamma,
            eff,
            effunc,
            effuncscale,
            exp,
            check_window,
        )

    # for iminuit ExtendedUnbinnedNLL
    def density(
        self,
        Es: np.array,
        S: float,
        BI: float,
        frac: float,
        delta: float,
        sigma: float,
        tau: float,
        gamma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        check_window: bool = False,
    ) -> np.array:
        return nb_density(
            Es,
            S,
            BI,
            frac,
            delta,
            sigma,
            tau,
            gamma,
            eff,
            effunc,
            effuncscale,
            exp,
            check_window,
        )

    # should we have an rvs method for drawing a random number of events?
    # `extendedrvs`
    # needs to use same parameters as the rest of the functions...
    def rvs(
        self,
        n_sig: int,
        n_bkg: int,
        frac: float,
        delta: float,
        sigma: float,
        tau: float,
        gamma: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_rvs(n_sig, n_bkg, frac, delta, sigma, tau, gamma, seed=seed)

    def extendedrvs(
        self,
        S: float,
        BI: float,
        frac: float,
        delta: float,
        sigma: float,
        tau: float,
        gamma: float,
        eff: float,
        effunc: float,
        effuncscale: float,
        exp: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_extendedrvs(
            S, BI, frac, delta, sigma, tau, gamma, eff, effunc, effuncscale, exp, seed=seed
        )

    # function call needs to take the same parameters as the other function calls, in the same order repeated twice
    # order is Es, S, BI, frac, delta, sigma, tau, gamma, eff, effunc, effuncscale, exp, check_window
    # this is intended only for empty datasets
    # returns `None` if we couldn't combine the datasets (a dataset was not empty)
    def combine(
        self,
        a_Es: np.array,
        a_S: float,
        a_BI: float,
        a_frac: float,
        a_delta: float,
        a_sigma: float,
        a_tau: float,
        a_gamma: float,
        a_eff: float,
        a_effunc: float,
        a_effuncscale: float,
        a_exp: float,
        b_Es: np.array,
        b_S: float,
        b_BI: float,
        b_frac: float,
        b_delta: float,
        b_sigma: float,
        b_tau: float,
        b_gamma: float,
        b_eff: float,
        b_effunc: float,
        b_effuncscale: float,
        b_exp: float,
    ) -> list | None:
        # datasets must be empty to be combined
        if len(a_Es) != 0 or len(b_Es) != 0:
            return None

        Es = np.array([])  # both of these datasets are empty
        S = 0.0  # this should be overwritten in the fit later
        BI = 0.0  # this should be overwritten in the fit later

        exp = a_exp + b_exp  # total exposure

        # exposure weighted fixed parameters (important to calculate correctly)
        sigma = (a_exp * a_sigma + b_exp * b_sigma) / exp
        eff = (a_exp * a_eff + b_exp * b_eff) / exp
        delta = (a_exp * a_delta + b_exp * b_delta) / exp
        tau = (a_exp * a_tau + b_exp * b_tau) / exp
        gamma = (a_exp * a_gamma + b_exp * b_gamma) / exp
        frac = (a_exp * a_frac + b_exp * b_frac) / exp

        # these are fully correlated in this model so the direct sum is appropriate
        # (maybe still appropriate even if not fully correlated?)
        effunc = (a_exp * a_effunc + b_exp * b_effunc) / exp

        effuncscale = 0.0  # this should be overwritten in the fit later

        return [Es, S, BI, frac, delta, sigma, tau, gamma, eff, effunc, effuncscale, exp]

    def can_combine(
        self,
        a_Es: np.array,
        a_S: float,
        a_BI: float,
        a_frac: float,
        a_delta: float,
        a_sigma: float,
        a_tau: float,
        a_gamma: float,
        a_eff: float,
        a_effunc: float,
        a_effuncscale: float,
        a_exp: float,
    ) -> bool:
        """
        This sets an arbitrary rule if this dataset can be combined with other datasets.
        In this case, if the dataset contains no data, then it can be combined, but more complex rules can be imposed.
        """
        if len(a_Es) == 0:
            return True
        else:
            return False


mjd_0vbb = mjd_0vbb_gen()



import numpy as np

QBB = 2039.0612  # 2039.0612 +- 0.0075 keV from AME2020
NA = 6.0221408e23  # Avogadro's number
M76 = 0.0759214027  # kilograms per mole, molar mass of 76Ge
MDET = 0.075681  # kg/mol, average molar mass for detectors from MJD Unidoc # M-TECHDOCUNIDOC-2022-068
G_01 = 0.23e-14  # in yr^-1, phase space factor for 76Ge from Phys. Rev. C 98, 035502 by Horoi et. al
g_A = 1.27  # noqa: N816 # axial coupling constant from Phys. Rev. Lett. 120, 202002 by Czarnecki et. al
me = 0.5109989500e6  # mass of the electron in eV/c^2 from PDG
NME_central = 2.60  # these values come from Phys. Rev. Lett. 132, 182502
NME_unc_hi = 1.28  # symmetrized error from Phys. Rev. Lett. 132, 182502
NME_unc_lo = 1.36  # symmetrized error from Phys. Rev. Lett. 132, 182502
# NME_PHENO_LOW = 2.12  # lower phenomenological  NME published after the US nuclear physics long range plan, Jokiniemi et al. https://doi.org/10.1103/PhysRevC.107.044305
NME_PHENO_LOW = 2.35  # lower phenomenological NME from the US nuclear physics long range plan arXiv:2212.11099, Jiao et al. doi.org/10.1103/PhysRevC.96.054310
NME_PHENO_HIGH = 6.34  # lower phenomenological NME from the US nuclear physics long range plan arXiv:2212.11099, Deppisch et al. doir.org/10.1103/PhysRevD.102.095016
# NME_PHENO_HIGH = 6.79  # upper phenomenological  NME published after the US nuclear physics long range plan, Jokiniemi et al. https://doi.org/10.1103/PhysRevC.107.044305

# lines to exclude are
# 2614.511(10) - 511 = 2103.511 keV SEP from 208Tl
# 2118.513(25) keV from 214Bi
# 2204.10(4) keV from 214Bi
# then round to nearest 0.1 keV

# default analysis window (in keV)
WINDOW = [[1930.0, 2098.5], [2108.5, 2113.5], [2123.5, 2190.0]]

# MJD analysis window (in keV) is slightly larger than GERDA/LEGEND and excludes an additional line
MJD_WINDOW = [
    [1950.0, 2098.5],
    [2108.5, 2113.5],
    [2123.5, 2199.1],
    [2209.1, 2350.0],
]

# could use these to go a little faster?
LOG2 = 0.69314718055994528622676398299518041312694549560546875
SQRT2PI = 2.506628274631000241612355239340104162693023681640625


# conversion function
def s_prime_to_s(s_prime):
    # Given s_prime in decays/(kg*yr), find s in decays/yr
    s = s_prime * (MDET / (LOG2 * NA))
    return s


# conversion function
def s_to_s_prime(s):
    # Given s in decays/yr, find s_prime in decays/(kg*yr)
    s_prime = s / (MDET / (LOG2 * NA))
    return s_prime


def halflife_to_s_prime(halflife):
    return s_to_s_prime(1.0 / halflife)


# another conversion function
def m_prime_to_m(m_prime):
    # Given m_prime in ev/(kg*yr), find m in ev
    return m_prime * np.sqrt(MDET * me**2 / (LOG2 * NA * G_01 * g_A**4))


def s_prime_to_halflife(s_prime):
    # Given s_prime in decays/(kg*yr), find t_half in yrs
    t_half = 1 / (s_prime_to_s(s_prime))
    return t_half



"""
Script for users to write their own initial guesses and pass them into the `experiment` and `toy` classes
"""

import numpy as np
from iminuit import Minuit

import freqfit.models.constants as constants
from freqfit.models.box_model_0vbb import box_model_0vbb_gen
from freqfit.models.correlated_efficiency_0vbb import (
    correlated_efficiency_0vbb_gen,
)
from freqfit.models.correlated_efficiency_0vbb_correlate_delta import (
    correlated_efficiency_0vbb_correlate_delta_gen,
)
from freqfit.models.correlated_efficiency_0vbb_exponential_background import (
    correlated_efficiency_0vbb_exponential_background_gen,
)
from freqfit.models.correlated_efficiency_0vbb_linear_background import (
    correlated_efficiency_0vbb_linear_background_gen,
)
from freqfit.models.correlated_efficiency_NME_0vbb import (
    correlated_efficiency_NME_0vbb_gen,
)
from freqfit.models.mjd_0vbb import mjd_0vbb_gen
from freqfit.models.mjd_NME_0vbb import mjd_NME_0vbb_gen
from freqfit.models.truncated_correlated_efficiency_0vbb import (
    truncated_correlated_efficiency_0vbb_gen,
)

# default analysis window and width
# window
#     uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
#     where edges of window are monotonically increasing (this is not checked), in keV.
#     Default is typical analysis window.
WINDOW = np.array(constants.WINDOW)

WINDOWSIZE = 0.0
for i in range(len(WINDOW)):
    WINDOWSIZE += WINDOW[i][1] - WINDOW[i][0]

QBB = constants.QBB


def zero_nu_initial_guess(experiment):
    # figure out if this is a toy or not:
    if hasattr(experiment, "experiment"):
        is_toy = True
        loop_exp = experiment.experiment
    else:
        is_toy = False
        loop_exp = experiment
    # Loop through the datasets and grab the exposures, efficiencies, and sigma from all datasets
    totexp = 0.0
    sigma_expweighted = 0.0
    eff_expweighted = 0.0
    effunc_expweighted = 0.0
    Es = []

    # Find which datasets share a background index
    BI_list = [par for par in experiment.fitparameters if "BI" in par]
    ds_list = []
    ds_names = []

    for BI in BI_list:
        ds_per_BI = []
        for ds in loop_exp.datasets.values():
            if is_toy:
                if (BI in ds.fitparameters) & (not ds._toy_is_combined):
                    ds_per_BI.append(ds)
            else:
                if (BI in ds.fitparameters) & (not ds.is_combined):
                    ds_per_BI.append(ds)

        if is_toy:
            for ds in experiment.combined_datasets.values():
                if BI in ds.fitparameters:
                    ds_per_BI.append(ds)
        else:
            for ds in loop_exp.combined_datasets.values():
                if BI in ds.fitparameters:
                    ds_per_BI.append(ds)

        ds_list.append(ds_per_BI)

        ds_names.append([ds.name for ds in ds_per_BI])

    # Fix all the fit parameters in the minuit object, then loosen S, all the BI and the global_effuncscale
    if is_toy:
        guess = {}
        for par in experiment.fitparameters:
            guess |= {
                par: experiment.experiment.toy_params_for_initial_guess[par]["value"]
            }
    else:
        guess = {
            fitpar: experiment.parameters[fitpar]["value"]
            if "value" in experiment.parameters[fitpar]
            else None
            for fitpar in experiment.fitparameters
        }

    minuit = Minuit(experiment.costfunction, **guess)
    for par in minuit.parameters:
        minuit.fixed[par] = True

    if "global_S" in list(minuit.fixed):
        minuit.fixed["global_S"] = False
        minuit.limits["global_S"] = (0, None)

    if "global_m_bb" in list(minuit.fixed):
        minuit.fixed["global_m_bb"] = False
        minuit.limits["global_m_bb"] = (0, None)

    # minuit.fixed["global_effuncscale"] = True
    # minuit.limits["global_effuncscale"] = (-100, 100)
    for BI in BI_list:
        minuit.fixed[f"{BI}"] = False
        minuit.limits[f"{BI}"] = (0, None)

        if "empty" in BI:
            minuit.fixed[f"{BI}"] = True

    # minuit.simplex()
    minuit.migrad()
    guess = minuit.values.to_dict()

    return guess


def guess_BI_S(Es, totexp, eff_expweighted, sigma_expweighted):  # noqa: N802
    """
    Give a better initial guess for the signal and background rate given an array of data
    The signal rate is estimated in a +/-5 keV window around Qbb, the BI is estimated from everything outside that window

    Parameters
    ----------
    Es
        A numpy array of observed energy data
    totexp
        The total exposure of the experiment
    eff_expweighted
        The total efficiency of the experiment
    sigma_expweighted
        The total sigma of the QBB peak
    """
    QBB_ROI_SIZE = [
        3 * sigma_expweighted,
        3 * sigma_expweighted,
    ]  # how many keV away from QBB in - and + directions we are defining the ROI
    BKG_WINDOW_SIZE = WINDOWSIZE - np.sum(
        QBB_ROI_SIZE
    )  # subtract off the keV we are counting as the signal region
    n_sig = 0
    n_bkg = 0
    for E in Es:
        if QBB - QBB_ROI_SIZE[0] <= E <= QBB + QBB_ROI_SIZE[1]:
            n_sig += 1
        else:
            n_bkg += 1

    # find the expected BI
    BI_guess = n_bkg / (BKG_WINDOW_SIZE * totexp)

    # # Now find the expected signal rate
    # n_sig -= (
    #     n_bkg * np.sum(QBB_ROI_SIZE) / BKG_WINDOW_SIZE
    # )  # subtract off the expected number of BI counts in ROI

    s_guess = n_sig / (totexp * eff_expweighted)

    return BI_guess, s_guess


def guess_BI_m_bb(Es, totexp, eff_expweighted, sigma_expweighted, NME):  # noqa: N802
    """
    Give a better initial guess for the signal and background rate given an array of data
    The signal rate is estimated in a +/-5 keV window around Qbb, the BI is estimated from everything outside that window

    Parameters
    ----------
    Es
        A numpy array of observed energy data
    totexp
        The total exposure of the experiment
    eff_expweighted
        The total efficiency of the experiment
    sigma_expweighted
        The total sigma of the QBB peak
    NME
        The central value of the NME used to generate this toy
    """
    QBB_ROI_SIZE = [
        3 * sigma_expweighted,
        3 * sigma_expweighted,
    ]  # how many keV away from QBB in - and + directions we are defining the ROI
    BKG_WINDOW_SIZE = WINDOWSIZE - np.sum(
        QBB_ROI_SIZE
    )  # subtract off the keV we are counting as the signal region
    n_sig = 0
    n_bkg = 0
    for E in Es:
        if QBB - QBB_ROI_SIZE[0] <= E <= QBB + QBB_ROI_SIZE[1]:
            n_sig += 1
        else:
            n_bkg += 1

    # find the expected BI
    BI_guess = n_bkg / (BKG_WINDOW_SIZE * totexp)

    m_guess = np.sqrt(n_sig / (totexp * eff_expweighted * NME**2))

    return BI_guess, m_guess


def poisson_initial_guess(experiment):
    # figure out if this is a toy or not:
    if hasattr(experiment, "experiment"):
        is_toy = True
        loop_exp = experiment.experiment
    else:
        is_toy = False
        loop_exp = experiment
    # Loop through the datasets and grab the exposures, efficiencies, and sigma from all datasets
    totexp = 0.0
    sigma_expweighted = 0.0
    eff_expweighted = 0.0
    effunc_expweighted = 0.0
    Es = []

    # Find which datasets share a background index
    BI_list = [par for par in experiment.fitparameters if "BI" in par]
    ds_list = []
    ds_names = []

    for BI in BI_list:
        ds_per_BI = []
        for ds in loop_exp.datasets.values():
            if is_toy:
                if (BI in ds.fitparameters) & (not ds._toy_is_combined):
                    ds_per_BI.append(ds)
            else:
                if (BI in ds.fitparameters) & (not ds.is_combined):
                    ds_per_BI.append(ds)

        if is_toy:
            for ds in experiment.combined_datasets.values():
                if BI in ds.fitparameters:
                    ds_per_BI.append(ds)
        else:
            for ds in loop_exp.combined_datasets.values():
                if BI in ds.fitparameters:
                    ds_per_BI.append(ds)

        ds_list.append(ds_per_BI)

        ds_names.append([ds.name for ds in ds_per_BI])

    # Fix all the fit parameters in the minuit object, then loosen S, all the BI and the global_effuncscale
    if is_toy:
        guess = {}
        for par in experiment.fitparameters:
            guess |= {
                par: experiment.experiment.toy_params_for_initial_guess[par]["value"]
            }
    else:
        guess = {
            fitpar: experiment.parameters[fitpar]["value"]
            if "value" in experiment.parameters[fitpar]
            else None
            for fitpar in experiment.fitparameters
        }

    # Then perform the loop over datasets that share a background index
    BI_guesses = []
    is_mbb_scan = False
    for ds_BI in ds_list:
        # Get estimates for these parameters based only on the datasets contributing to one BI
        BI_totexp = 0.0
        BI_sigma_expweighted = 0.0
        BI_eff_expweighted = 0.0
        BI_effunc_expweighted = 0.0
        Es_per_BI = []

        # need to handle this differently depending on if this is a toy or not/what datasets have been combined
        for ds in ds_BI:
            if (
                isinstance(ds.model, correlated_efficiency_0vbb_gen)
                or isinstance(ds.model, box_model_0vbb_gen)
                or isinstance(ds.model, truncated_correlated_efficiency_0vbb_gen)
                or isinstance(ds.model, correlated_efficiency_NME_0vbb_gen)
                or isinstance(ds.model, correlated_efficiency_0vbb_correlate_delta_gen)
                or isinstance(
                    ds.model, correlated_efficiency_0vbb_linear_background_gen
                )
                or isinstance(
                    ds.model, correlated_efficiency_0vbb_exponential_background_gen
                )
            ):
                BI_totexp = BI_totexp + ds._parlist[7]
                BI_sigma_expweighted = (
                    BI_sigma_expweighted + ds._parlist[3] * ds._parlist[7]
                )
                BI_eff_expweighted = (
                    BI_eff_expweighted + ds._parlist[4] * ds._parlist[7]
                )
                BI_effunc_expweighted = (
                    BI_effunc_expweighted + ds._parlist[5] * ds._parlist[7]
                )
                if is_toy:
                    if ds._toy_data is not None:
                        Es_per_BI.extend(ds._toy_data)
                    else:
                        Es_per_BI.extend([])
                else:
                    Es_per_BI.extend(ds.data)
                # report the NME if this is an m_bb toy
                if isinstance(ds.model, correlated_efficiency_NME_0vbb_gen):
                    NME = ds._parlist[-1]
                    is_mbb_scan = True  # we cannot mix and match S and m_bb, if this is true we are using m_bb only
            elif isinstance(ds.model, mjd_0vbb_gen) or isinstance(
                ds.model, mjd_NME_0vbb_gen
            ):
                BI_totexp = BI_totexp + ds._parlist[10]
                BI_sigma_expweighted = (
                    BI_sigma_expweighted + ds._parlist[4] * ds._parlist[10]
                )
                BI_eff_expweighted = (
                    BI_eff_expweighted + ds._parlist[7] * ds._parlist[10]
                )
                BI_effunc_expweighted = (
                    BI_effunc_expweighted + ds._parlist[8] * ds._parlist[10]
                )
                if is_toy:
                    if ds._toy_data is not None:
                        Es_per_BI.extend(ds._toy_data)
                    else:
                        Es_per_BI.extend([])
                else:
                    Es_per_BI.extend(ds.data)
            else:
                raise NotImplementedError(
                    f"Model of type {ds.model} not yet implemented here!"
                )

        totexp += BI_totexp
        sigma_expweighted += BI_sigma_expweighted
        eff_expweighted += BI_eff_expweighted
        effunc_expweighted += BI_effunc_expweighted
        Es.extend(Es_per_BI)

        BI_sigma_expweighted = BI_sigma_expweighted / BI_totexp
        BI_eff_expweighted = BI_eff_expweighted / BI_totexp
        BI_effunc_expweighted = BI_effunc_expweighted / BI_totexp

        # Finally, we are ready to make our guess for this BI
        # If we get only one count in the signal window, then this guess will estimate too low a background
        # So, if BI is guessed as 0 and S is not 0, smear out the signal rate between them
        if is_mbb_scan:
            BI_guess, m_bb_guess = guess_BI_m_bb(
                Es_per_BI, BI_totexp, BI_eff_expweighted, BI_sigma_expweighted, NME
            )
            if m_bb_guess < 0:
                m_bb_guess = 0

            if BI_guess <= 0:
                BI_guess = 0
        else:
            BI_guess, s_guess = guess_BI_S(
                Es_per_BI, BI_totexp, BI_eff_expweighted, BI_sigma_expweighted
            )
            if s_guess < 0:
                s_guess = 0

            if BI_guess <= 0:
                BI_guess = 0

        BI_guesses.append(BI_guess)

    # Compute the total for the experiment, so that we can better guess an initial S value
    sigma_expweighted = sigma_expweighted / BI_totexp
    eff_expweighted = eff_expweighted / BI_totexp
    effunc_expweighted = effunc_expweighted / BI_totexp

    if is_mbb_scan:
        m_bb_guess = guess_BI_m_bb(Es, totexp, eff_expweighted, sigma_expweighted, NME)[
            1
        ]
        if m_bb_guess <= 0:
            m_bb_guess = 0
    else:
        S_guess = guess_BI_S(Es, totexp, eff_expweighted, sigma_expweighted)[1]
        if S_guess <= 0:
            S_guess = 0

    # need to handle this differently depending on if this is a toy or not/what datasets have been combined
    if is_toy:
        guess = {}
        for par in experiment.fitparameters:
            guess |= {
                par: experiment.experiment.toy_params_for_initial_guess[par]["value"]
            }
    else:
        guess = {
            fitpar: experiment.parameters[fitpar]["value"]
            if "value" in experiment.parameters[fitpar]
            else None
            for fitpar in experiment.fitparameters
        }

    # update the BI
    for i, BI in enumerate(BI_list):
        guess[BI] = BI_guesses[i]

    # Update the signal guess
    if is_mbb_scan:
        guess["global_m_bb"] = m_bb_guess
    else:
        guess["global_S"] = S_guess

    return guess



"""
A class that holds a collection of fake datasets and associated hardware
"""
import itertools
import logging
from copy import deepcopy

import numpy as np
from iminuit import Minuit, cost

from freqfit import initial_guesses
from freqfit.dataset import combine_datasets
from freqfit.utils import grab_results

SEED = 42

log = logging.getLogger(__name__)


class Toy:
    def __init__(
        self,
        experiment,
        parameters: dict,
        seed: int = SEED,
    ) -> None:
        """
        experiment
            `experiment` to base this `Toy` on
        parameters
            `dict` of parameters and their values to model with
        """

        self.experiment = experiment  # the experiment this Toy is based on
        self.toy_data_to_save = (
            []
        )  # list to store the data drawn for this toy. A flat array, consists of all data from all datasets
        self.toy_num_drawn_to_save = (
            []
        )  # list to store tuples of the number of signal and background counts drawn per dataset
        self.parameters_to_save = np.array(
            [np.full(len(self.experiment._toy_parameters), np.nan)]
        )  # the parameters of the toy
        self.costfunction = None  # costfunction for this Toy
        self.fitparameters = None  # fit parameters from the costfunction, reference to self.costfunction._parameters
        self.minuit = None  # Minuit object
        self.iminuit_tolerance = self.experiment.iminuit_tolerance
        self.iminuit_precision = self.experiment.iminuit_precision
        self.minimizer_options = self.experiment.minimizer_options
        self.iminuit_strategy = self.experiment.iminuit_strategy
        self.guess = None  # initial guess for minuit
        self.best = None  # best fit
        self.fixed_bc_no_data = (
            {}
        )  # parameters that can be fixed because no data in their Datasets
        self.combined_datasets = {}  # holds combined_datasets
        self.included_in_combined_datasets = {}
        self.seed = seed
        self.user_gradient = self.experiment.user_gradient
        self.use_log = self.experiment.use_log
        self.scan = self.experiment.scan
        self.scan_grid = self.experiment.scan_grid
        self.data = (
            []
        )  # A flat array of all the data. The data may be split between datasets, this is just an aggregate

        # reset toy_parameters to "default" values
        # deepcopy so we can mutate this without fear
        self.experiment._toy_parameters = deepcopy(self.experiment.parameters)

        # overwrite the toy parameters with the passed parameters
        for par in parameters.keys():
            self.experiment._toy_parameters[par]["value"] = parameters[par]

        # Initial guesses for the toy fit need to recenter the drawn nuisance parameters, and need to be in the correct order
        self.experiment.toy_params_for_initial_guess = deepcopy(
            self.experiment._toy_parameters
        )

        # If datasets have been combined, re-assign those combined parameter values to the de-combined toy_parameters
        for combined_ds in experiment.included_in_combined_datasets.keys():
            # Datasets may have been combined into more than one new datasets, that's why we loop over the keys
            # Now, loop over datasets and add back in their parameter values
            for ds in experiment.included_in_combined_datasets[combined_ds]:
                # Check to make sure that the user hasn't overridden any of the parameters in datasets that were combined
                # This needs to happen so we don't overwrite them when we de-combine datasets next
                if any(
                    x in list(parameters.keys())
                    for x in list(experiment.datasets[ds].model_parameters.keys())
                ):
                    raise NotImplementedError(
                        "Overriding a parameter that is in a combined dataset is not supported."
                    )

                # Loop through the model parameters in the dataset and find their values from the combined dataset
                for model_par in self.experiment.datasets[ds].model_parameters.keys():
                    ds_par_name = self.experiment.datasets[ds].model_parameters[
                        model_par
                    ]
                    # find the corresponding parameter name in the combined dataset
                    combined_ds_par_name = self.experiment.combined_datasets[
                        combined_ds
                    ].model_parameters[model_par]

                    # nuisance parameters should not be passed in parameters, so we need to skip the nuisance parameters from the dataset
                    if combined_ds_par_name in parameters.keys():
                        parameters[ds_par_name] = parameters[combined_ds_par_name]

        # vary the toy parameters as indicated
        if len(self.experiment._toy_pars_to_vary) > 0:
            np.random.seed(seed=self.seed)

            pars, vals, covar = self.experiment.get_constraints(
                self.experiment._toy_pars_to_vary
            )

            varied_toy_pars = []
            # check if parameters are all independent, draw from simpler distribution if so
            if np.all(covar == np.diag(np.diagonal(covar))):
                varied_toy_pars = np.random.normal(vals, np.sqrt(np.diagonal(covar)))
            else:
                varied_toy_pars = np.random.multivariate_normal(
                    vals, covar
                )  # sooooooooo SLOW

            # Check that toypars are drawn within the parameter limits here
            check_toy_params_beyond_limits = True
            check_toy_params_beyond_limits_counter = 0  # pressure relief valve
            while check_toy_params_beyond_limits & (
                check_toy_params_beyond_limits_counter < 100
            ):
                check_toy_params_beyond_limits_counter += 1
                if check_toy_params_beyond_limits_counter == 100:
                    logging.warning(
                        "exiting toy parameter redrawing as maximum number of redraws has been reached."
                    )

                for i, par in enumerate(pars):
                    par_lims = [None, None]
                    if "limits" in self.experiment._toy_parameters[par]:
                        par_lims = self.experiment._toy_parameters[par]["limits"]

                    # check if we are within the limits
                    if (
                        (par_lims[0] if par_lims[0] is not None else -np.inf)
                        <= varied_toy_pars[i]
                        <= (par_lims[1] if par_lims[1] is not None else np.inf)
                    ):
                        check_toy_params_beyond_limits = False
                    # Parameter does not fall within its limit, redraw them and rerun the loop
                    else:
                        logging.warning(
                            "redrawing toy values because drawn value falls outside allowed limit"
                        )
                        check_toy_params_beyond_limits = True
                        # check if parameters are all independent, draw from simpler distribution if so
                        if np.all(covar == np.diag(np.diagonal(covar))):
                            varied_toy_pars = np.random.normal(
                                vals, np.sqrt(np.diagonal(covar))
                            )
                        else:
                            varied_toy_pars = np.random.multivariate_normal(
                                vals, covar
                            )  # sooooooooo SLOW
                        break

            # now assign the random values to the passed parameters (or to not passed parameters?)
            for i, par in enumerate(pars):
                parameters[par] = varied_toy_pars[i]
                self.experiment.toy_params_for_initial_guess[par][
                    "value"
                ] = varied_toy_pars[i]
                # commented out 01/22/2025 -- toys should be generated at nuisance parameters nominal value, but the constraints should be re-centered to the randomly drawn value, replaced with line below
                # self.experiment._toy_parameters[par]["value"] = varied_toy_pars[i]

        # draw the toy data
        for i, (datasetname, dataset) in enumerate(self.experiment.datasets.items()):
            # worried that this is not totally deterministic (depends on number of Datasets),
            # but more worried that random draws between datasets would be correlated otherwise.
            thisseed = self.seed + i

            # allocate length in case `parameters` is passed out of order
            pars = [None for j in range(len(dataset.fitparameters))]

            for j, (fitpar, fitparindex) in enumerate(dataset.fitparameters.items()):
                if fitpar not in parameters:
                    msg = f"`Toy`: for `Dataset` {datasetname}, parameter `{fitpar}` not found in passed `parameters`"
                    raise KeyError(msg)

                # commented out 01/22/2025 -- toys should be generated at nuisance parameters nominal value, but the constraints should be re-centered to the randomly drawn value, replaced with line below
                # pars[j] = parameters[fitpar]
                pars[j] = self.experiment._toy_parameters[fitpar]["value"]

            # make the toy for this particular dataset
            dataset.toy(par=pars, seed=thisseed)  # saved in dataset._toy_data
            self.toy_data_to_save.extend(
                dataset._toy_data
            )  # store the data as list of lists
            self.toy_num_drawn_to_save.extend(
                dataset._toy_num_drawn
            )  # also save the number of signal and background counts drawn

        # combine datasets
        if self.experiment.try_to_combine_datasets:
            # maybe there's more than one combined_dataset group
            for cdsname in self.experiment._combined_datasets_config:
                # find the Datasets to try to combine
                ds_tocombine = []
                for dsname in self.experiment.datasets.keys():
                    if (
                        self.experiment.datasets[dsname].try_combine
                        and self.experiment.datasets[dsname].combined_dataset == cdsname
                    ):
                        ds_tocombine.append(self.experiment.datasets[dsname])

                combined_dataset, included_datasets = combine_datasets(
                    datasets=ds_tocombine,
                    model=self.experiment._combined_datasets_config[cdsname]["model"],
                    model_parameters=self.experiment._combined_datasets_config[cdsname][
                        "model_parameters"
                    ],
                    parameters=self.experiment._toy_parameters,
                    costfunction=self.experiment._combined_datasets_config[cdsname][
                        "costfunction"
                    ],
                    name=cdsname,
                    use_toy_data=True,
                    user_gradient=self.user_gradient,
                    use_log=self.use_log,
                )

                # now to see what datasets actually got included in the combined_dataset
                for dsname in included_datasets:
                    self.experiment.datasets[dsname]._toy_is_combined = True

                if len(included_datasets) > 0:
                    self.combined_datasets[cdsname] = combined_dataset
                    self.included_in_combined_datasets[cdsname] = included_datasets

        # to find which parameters are part of Datasets that have data
        parstofitthathavedata = set()

        # add the costfunctions together
        first = True
        for dsname, ds in self.experiment.datasets.items():
            # skip this dataset if it has been combined
            if ds._toy_is_combined:
                continue
            # make the cost function for this particular dataset
            if self.use_log:
                thiscostfunction = ds._costfunctioncall(
                    ds._toy_data, ds.log_density, log=True
                )
            else:
                thiscostfunction = ds._costfunctioncall(ds._toy_data, ds.density)
            # tell the cost function which parameters to use
            thiscostfunction._parameters = ds.costfunction._parameters
            if first:
                self.costfunction = thiscostfunction
                first = False
            else:
                self.costfunction += thiscostfunction

            # find which parameters are part of Datasets that have data
            if ds._toy_data.size > 0:
                # Add the data to self.data
                self.data.extend(ds._toy_data)
                # add the fit parameters of this Dataset if there is some data
                for fitpar in thiscostfunction._parameters:
                    parstofitthathavedata.add(fitpar)

        for cdsname, cds in self.combined_datasets.items():
            # make the cost function for this particular combined_dataset
            if self.use_log:
                thiscostfunction = cds._costfunctioncall(
                    cds.data, cds.log_density, log=True
                )
            else:
                thiscostfunction = cds._costfunctioncall(cds.data, cds.density)
            # tell the cost function which parameters to use
            thiscostfunction._parameters = cds.costfunction._parameters
            if first:
                self.costfunction = thiscostfunction
                first = False
            else:
                self.costfunction += thiscostfunction

            # find which parameters are part of combined_datasets that have data
            if cds.data.size > 0:
                # add the fit parameters of this combined_dataset if there is some data
                for fitpar in thiscostfunction._parameters:
                    parstofitthathavedata.add(fitpar)

        # fitparameters of Toy are a little different than fitparameters of Dataset
        self.fitparameters = self.costfunction._parameters

        # get the constraints needed from the fitparameters
        # but note that we need to adjust the central values for the varied pars
        # check if they are in the varied pars list and pull the value from parameters if so

        pars, values, covariance = self.experiment.get_constraints(self.fitparameters)

        if pars is not None:
            for i, par in enumerate(pars):
                if par in self.experiment._toy_pars_to_vary:
                    values[i] = parameters[par]

            self.constraints = cost.NormalConstraint(pars, values, error=covariance)

            self.costfunction = self.costfunction + self.constraints

        self.guess = self.initialguess()

        self.minuit = Minuit(self.costfunction, **self.guess)
        self.minuit.tol = self.iminuit_tolerance  # set the tolerance
        self.minuit.precision = self.iminuit_precision
        self.minuit.strategy = self.iminuit_strategy

        # raise a RunTime error if function evaluates to NaN
        self.minuit.throw_nan = True

        # now check which nuisance parameters can be fixed if no data and are not part of a Dataset that has data
        for parname in self.fitparameters:
            if (
                "fix_if_no_data" in self.experiment._toy_parameters[parname]
                and self.experiment._toy_parameters[parname]["fix_if_no_data"]
            ):
                # check if this parameter is part of a Dataset that has data
                if parname not in parstofitthathavedata:
                    self.fixed_bc_no_data[parname] = self.experiment._toy_parameters[
                        parname
                    ]["value"]

        # save the values of the toy parameters
        for i, (parname, pardict) in enumerate(
            self.experiment.toy_params_for_initial_guess.items()
        ):
            self.parameters_to_save[0, i] = pardict["value"]

        # to set limits and fixed variables
        # this function will also fix those nuisance parameters which can be fixed because they are not part of a
        # Dataset that has data
        self.minuit_reset()

        # If scan is not none, initialize the hypercube grid
        if (self.scan) and (self.scan_grid is not None):
            self.hypercube_grid = self.create_hypercube(self.scan_grid)

    def initialguess(
        self,
    ) -> dict:
        if self.experiment.initial_guess_function is None:
            guess = {}
            for par in self.fitparameters:
                guess |= {
                    par: self.experiment.toy_params_for_initial_guess[par]["value"]
                }

        else:
            func = getattr(initial_guesses, self.experiment.initial_guess_function)
            guess = func(self)

        return guess

    def minuit_reset(
        self,
        use_physical_limits: bool = True,  # for numerators of test statistics, want this to be False
    ) -> None:
        # resets the minimization and stuff
        # does not change limits but does remove "fixed" attribute of variables
        # 2024/08/09: This no longer seems to be true? Not sure if something changed in iminuit or if I was wrong?
        self.minuit.reset()

        # Restore the first initial guess
        self.minuit.values[:] = list(self.guess.values())

        # overwrite the limits
        # note that this information can also be contained in a Dataset when instantiated
        # and is overwritten here

        # also set which parameters are fixed
        for parname, pardict in self.experiment._toy_parameters.items():
            if parname in self.minuit.parameters:
                self.minuit.fixed[parname] = False
                self.minuit.limits[parname] = (-1.0 * np.inf, np.inf)

                if "limits" in pardict:
                    self.minuit.limits[parname] = pardict["limits"]
                if use_physical_limits and "physical_limits" in pardict:
                    self.minuit.limits[parname] = pardict["physical_limits"]
                if "fixed" in pardict:
                    self.minuit.fixed[parname] = pardict["fixed"]
                # fix those nuisance parameters which can be fixed because they are not part of a
                # Dataset that has data
                if parname in self.fixed_bc_no_data:
                    self.minuit.fixed[parname] = True
                    self.minuit.values[parname] = self.fixed_bc_no_data[parname]

        return

    def bestfit(
        self,
        force: bool = False,
        use_physical_limits: bool = True,
    ) -> dict:
        # don't run this more than once if we don't have to
        if self.best is not None and not force:
            return self.best

        # remove any previous minimizations
        self.minuit_reset(use_physical_limits=use_physical_limits)

        if self.scan:
            y = np.empty(len(self.hypercube_grid))
            for i in range(len(self.hypercube_grid)):
                y[i] = self.minuit._fcn(self.hypercube_grid[i])

            best = np.min(y)
            ibest = np.argmin(y)

            self.best = {
                "fval": best,
                "values": {},
                "valid": True,
            }  # TODO: add the rest of return values of interest, like parameter names at minimum
            for par in self.minuit.parameters:
                ipar, vpar = self.minuit._normalize_key(par)
                self.best["values"][par] = self.hypercube_grid[ibest][ipar]
        else:
            try:
                if self.experiment.backend == "minuit":
                    self.minuit.migrad(**self.minimizer_options)
                elif self.experiment.backend == "scipy":
                    self.minuit.scipy(
                        method=self.experiment.scipy_minimizer,
                        options=self.minimizer_options,
                    )
                else:
                    raise NotImplementedError(
                        "Iminuit backend is not set to `minuit` or `scipy`"
                    )

            except RuntimeError:
                msg = f"`Toy` with seed {self.seed} has best fit throwing NaN"
                logging.warning(msg)

            if not self.minuit.valid:
                msg = f"`Toy` with seed {self.seed} has invalid best fit"
                logging.debug(msg)

            self.best = grab_results(
                self.minuit,
                use_grid_rounding=self.experiment.use_grid_rounding,
                grid_rounding_num_decimals=self.experiment.grid_rounding_num_decimals,
            )

        if "global_S" in self.best["values"]:
            if self.best["values"]["global_S"] < 1e-20:
                self.best = self.profile({"global_S": 0.0})

        if "global_m_bb" in self.best["values"]:
            if self.best["values"]["global_m_bb"] < 1e-20:
                self.best = self.profile({"global_m_bb": 0.0})

        if self.guess == self.best["values"]:
            msg = f"`Toy` with seed {self.seed} has best fit values very close to initial guess"
            logging.debug(msg)

        return self.best

    def profile(
        self,
        parameters: dict,
        use_physical_limits: bool = True,
    ) -> dict:
        """
        parameters
            `dict` where keys are names of parameters to fix and values are the value that the parameter should be
            fixed to
        """

        self.minuit_reset(use_physical_limits=use_physical_limits)

        for parname, parvalue in parameters.items():
            self.minuit.fixed[parname] = True
            self.minuit.values[parname] = parvalue

        if self.scan:
            # need to remake the hypercube because there are now fewer parameters
            # pop the fixed parameters from the supplied scan_grid
            new_grid_dict = deepcopy(self.scan_grid)
            for parname in parameters.keys():
                if parname in new_grid_dict.keys():
                    new_grid_dict.pop(parname)
            hypercube_grid = self.create_hypercube(new_grid_dict)

            y = np.empty(len(hypercube_grid))
            for i in range(len(hypercube_grid)):
                y[i] = self.minuit._fcn(hypercube_grid[i])

            best = np.min(y)
            ibest = np.argmin(y)
            results = {
                "fval": best,
                "values": {},
                "valid": True,
            }  # TODO: add the rest of return values of interest
            for par in self.minuit.parameters:
                ipar, vpar = self.minuit._normalize_key(par)
                results["values"][par] = hypercube_grid[ibest][ipar]

        else:
            try:
                if self.experiment.backend == "minuit":
                    self.minuit.migrad(**self.minimizer_options)
                elif self.experiment.backend == "scipy":
                    self.minuit.scipy(
                        method=self.experiment.scipy_minimizer,
                        options=self.minimizer_options,
                    )
                else:
                    raise NotImplementedError(
                        "Iminuit backend is not set to `minuit` or `scipy`"
                    )
            except RuntimeError:
                msg = f"`Toy` with seed {self.seed} has profile raising NaN with parameters {parameters}"
                logging.warning(msg)

            if not self.minuit.valid:
                msg = f"`Toy` with seed {self.seed} has invalid profile"
                logging.debug(msg)

            results = grab_results(
                self.minuit,
                use_grid_rounding=self.experiment.use_grid_rounding,
                grid_rounding_num_decimals=self.experiment.grid_rounding_num_decimals,
            )

        if self.guess == results["values"]:
            msg = f"`Toy` with seed {self.seed} has profile fit values very close to initial guess"
            logging.debug(msg)

        # also include the fixed parameters
        for parname, parvalue in parameters.items():
            results[parname] = parvalue

        return results

    # this corresponds to t_mu or t_mu^tilde depending on whether there is a physical limit on the parameters
    def ts(
        self,
        profile_parameters: dict,  # which parameters to fix and their value (rest are profiled)
        force: bool = False,
    ) -> float:
        """
        force
            See `experiment.bestfit()` for description. Default is `False`.
        """

        use_physical_limits = False  # for t_mu and q_mu
        if (
            self.experiment.test_statistic == "t_mu_tilde"
            or self.experiment.test_statistic == "q_mu_tilde"
        ):
            use_physical_limits = True

        denom = self.bestfit(force=force, use_physical_limits=use_physical_limits)[
            "fval"
        ]

        # see Cowan (2011) Eq. 14 and Eq. 16
        if (
            self.experiment.test_statistic == "q_mu"
            or self.experiment.test_statistic == "q_mu_tilde"
        ):
            for parname, parvalue in profile_parameters.items():
                if self.best["values"][parname] > parvalue:
                    return 0.0, 0.0, 0.0

        num = self.profile(
            parameters=profile_parameters, use_physical_limits=use_physical_limits
        )["fval"]

        ts = num - denom

        if ts < 0:
            # update initial guess and try again
            for par in list(self.guess.keys()):
                self.guess[par] = self.minuit.values[par]
            denom = self.bestfit(force=force, use_physical_limits=use_physical_limits)[
                "fval"
            ]
            num = self.profile(
                parameters=profile_parameters, use_physical_limits=use_physical_limits
            )["fval"]

            ts = num - denom

            if ts < 0:
                msg = f"`Toy` profiled at {profile_parameters} with seed {self.seed} gave test statistic below zero: {ts}"
                logging.warning(msg)

        # because these are already -2*ln(L) from iminuit
        return ts, denom, num

    # mostly pulled directly from iminuit, with some modifications to ignore empty Datasets and also to format
    # plots slightly differently
    def visualize(
        self,
        component_kwargs=None,
    ) -> None:
        """
        Visualize the last toy.

        The visualization is drawn with matplotlib.pyplot into the current figure.
        Subplots are created to visualize each part of the cost function, the figure
        height is increased accordingly. Parts without a visualize method are silently
        ignored.

        Does not draw Datasets that are empty (have no events).

        Parameters
        ----------
        args : array-like
            Parameter values.
        component_kwargs : dict of dicts, optional
            Dict that maps an index to dict of keyword arguments. This can be
            used to pass keyword arguments to a visualize method of a component with
            that index.
        **kwargs :
            Other keyword arguments are forwarded to all components.
        """
        from matplotlib import pyplot as plt

        parameters = {}
        for par, pardict in self.experiment._toy_parameters.items():
            parameters[par] = pardict["value"]

        args = []
        for par in self.fitparameters:
            if par not in parameters:
                msg = f"parameter {par} was not provided"
                raise KeyError(msg)
            args.append(parameters[par])

        n = 0
        for comp in self.costfunction:
            if hasattr(comp, "visualize"):
                if hasattr(comp, "data"):
                    if len(comp.data) > 0:
                        n += 1
                else:
                    n += 1

        fig = plt.gcf()
        fig.set_figwidth(n * fig.get_figwidth() / 1.5)
        _, ax = plt.subplots(1, n, num=fig.number)

        if component_kwargs is None:
            component_kwargs = {}

        i = 0
        for k, (comp, cargs) in enumerate(self.costfunction._split(args)):
            if not hasattr(comp, "visualize"):
                continue
            if hasattr(comp, "data") and len(comp.data) == 0:
                continue
            kwargs = component_kwargs.get(k, {})
            plt.sca(ax[i])
            comp.visualize(cargs, **kwargs)
            i += 1

    def create_hypercube(self, scan_grid) -> list:
        """
        Parameters
        ----------
        m
            Minuit object that holds that parameters and parameter names
        scan_grid
            The dictionary object that holds the points in each dimension to scan
        """
        scan_pars = list(scan_grid.keys())
        scan_list_of_lists = [list(scan_grid[scan_par]) for scan_par in scan_pars]
        hypercube_missing_other_values = itertools.product(
            *scan_list_of_lists
        )  # This list is missing all of the other function values that the minuit object needs, let's add them back in!

        # get the indices within the minuit object of all of parameters on the grid
        ipar_list = []
        for scan_par in scan_pars:
            ipar_list.append(self.minuit._normalize_key(scan_par)[0])
        ipar_list = np.array(ipar_list)
        values = deepcopy(
            np.array(self.minuit.values)
        )  # Get all the values, don't mutate them

        hypercube = []
        for par in hypercube_missing_other_values:
            values[ipar_list] = par
            hypercube.append(deepcopy(values))

        return hypercube



import numpy as np
from iminuit import Minuit, cost

import freqfit.models.constants as constants
from freqfit.models import gaussian_on_uniform

QBB = constants.QBB
N_A = constants.NA
M76 = constants.M76
WINDOW = np.array(constants.WINDOW)

WINDOWSIZE = 0.0
for i in range(len(WINDOW)):
    WINDOWSIZE += WINDOW[i][1] - WINDOW[i][0]


def test_gaussian_on_uniform_pdf():
    eff = 0.6
    exp = 100
    S = 1e-24
    BI = 2e-4
    delta = 0
    sigma = 1

    Es = np.linspace(1930, 2190, 100)

    # compute using our function
    model_values = gaussian_on_uniform.pdf(Es, S, BI, delta, sigma, eff, exp)

    # Compute using scipy
    mu_S = S * eff * exp
    mu_B = exp * BI * WINDOWSIZE
    scipy_values = (1 / (mu_S + mu_B)) * (
        mu_S
        * np.exp(-((Es - QBB - delta) ** 2) / (2 * (sigma**2)))
        / (np.sqrt(2 * np.pi) * sigma)
        + mu_B / WINDOWSIZE
    )

    assert np.allclose(model_values, scipy_values, rtol=1e-3)


def test_gaussian_on_uniform_logpdf():
    eff = 0.6
    exp = 100
    S = 1e-24
    BI = 2e-4
    delta = 0
    sigma = 1

    Es = np.linspace(1930, 2190, 100)

    # compute using our function
    model_values = gaussian_on_uniform.logpdf(Es, S, BI, delta, sigma, eff, exp)

    # Compute using scipy
    mu_S = S * eff * exp
    mu_B = exp * BI * WINDOWSIZE
    scipy_values = np.log(
        (1 / (mu_S + mu_B))
        * (
            mu_S
            * np.exp(-((Es - QBB - delta) ** 2) / (2 * (sigma**2)))
            / (np.sqrt(2 * np.pi) * sigma)
            + mu_B / WINDOWSIZE
        )
    )

    assert np.allclose(model_values, scipy_values, rtol=1e-3)


def test_gaussian_on_uniform_rvs():
    n_sig = 1000
    n_bkg = 100
    delta = 0
    sigma = 1

    random_sample = gaussian_on_uniform.rvs(n_sig, n_bkg, delta, sigma)

    assert np.allclose(np.median(random_sample), QBB, rtol=1e-1)


def test_iminuit_integration():
    n_sig = 1000
    n_bkg = 100
    delta = 0
    sigma = 1

    random_sample = gaussian_on_uniform.rvs(n_sig, n_bkg, delta, sigma)

    c = cost.UnbinnedNLL(random_sample, gaussian_on_uniform.pdf)
    m = Minuit(c, S=0.5, BI=0.1, delta=-0.1, sigma=0.6, eff=0.9, exp=0.9)
    m.fixed["eff", "exp"] = True
    m.limits["sigma", "S", "BI"] = (0, None)
    m.migrad()

    assert np.allclose(m.values["sigma"], 1, rtol=1e-1)
    assert np.allclose(m.values["delta"], 0.01, rtol=1e0)

    c = cost.UnbinnedNLL(random_sample, gaussian_on_uniform.logpdf, log=True)
    m = Minuit(c, S=0.5, BI=0.1, delta=-0.1, sigma=0.6, eff=0.9, exp=0.9)
    m.fixed["eff", "exp"] = True
    m.limits["sigma", "S", "BI"] = (0, None)
    m.migrad()

    assert np.allclose(m.values["sigma"], 1, rtol=1e-1)
    assert np.allclose(m.values["delta"], 0.01, rtol=1e0)


def test_density_gradient():
    n_sig = 1000
    n_bkg = 100
    delta = 0
    sigma = 1

    random_sample = gaussian_on_uniform.rvs(n_sig, n_bkg, delta, sigma)

    c = cost.ExtendedUnbinnedNLL(
        random_sample,
        gaussian_on_uniform.density,
        grad=gaussian_on_uniform.density_gradient,
    )
    m = Minuit(c, S=1.0, BI=0.1, delta=-0.1, sigma=0.6, eff=0.9, exp=0.9)
    m.fixed["eff", "exp"] = True
    m.limits["sigma", "S", "BI"] = (0, None)
    m.migrad()

    assert np.allclose(m.values["sigma"], 1, rtol=1e-1)
    assert np.allclose(m.values["delta"], 0.01, rtol=1e0)


def test_logdensity():
    n_sig = 1000
    n_bkg = 100
    delta = 0
    sigma = 1

    random_sample = gaussian_on_uniform.rvs(n_sig, n_bkg, delta, sigma)

    c = cost.ExtendedUnbinnedNLL(
        random_sample, gaussian_on_uniform.log_density, log=True
    )
    m = Minuit(c, S=1, BI=0.1, delta=-0.1, sigma=0.6, eff=0.9, exp=0.9)
    m.limits["sigma", "S", "BI"] = (0, None)
    m.fixed["eff", "exp"] = True
    m.migrad()

    assert np.allclose(m.values["sigma"], 1, rtol=1e-1)
    assert np.allclose(m.values["delta"], 0.01, rtol=1e0)



# Configuration file for the Sphinx documentation builder.

import sys
from pathlib import Path

from pkg_resources import get_distribution

sys.path.insert(0, Path("../../src").resolve().as_posix())
sys.path.append(Path("extensions").resolve().as_posix())

project = "=freqfit"
copyright = "2024, the UW LEGEND Group on Behalf of the LEGEND Collaboration"
version = get_distribution("=freqfit").version

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx_copybutton",
    "sphinx_inline_tabs",
    "nbsphinx",
    "myst_parser",
    "IPython.sphinxext.ipython_console_highlighting",
    "numbadoc",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}
master_doc = "index"

# Furo theme
html_theme = "furo"
html_theme_options = {
    "source_repository": "https://github.com/cenpa/=freqfit",
    "source_branch": "main",
    "source_directory": "docs/source",
}
html_title = f"{project} {version}"
html_logo = "../../.github/logo.png"

# sphinx-napoleon
# enforce consistent usage of NumPy-style docstrings
napoleon_numpy_docstring = True
napoleon_google_docstring = False
napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_custom_sections = ["JSON Configuration Example"]


# intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "numba": ("https://numba.readthedocs.io/en/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "pandas": ("https://pandas.pydata.org/docs", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "h5py": ("https://docs.h5py.org/en/stable", None),
    "pint": ("https://pint.readthedocs.io/en/stable", None),
}

suppress_warnings = [
    # "histogram" is defined both in pygama.dsp.processors.histogram.histogram
    # and in pygama.math.histogram, leading to a Sphinx cross-referencing
    # warning.  I don't know how to fix this properly
    "ref.python",
]

# sphinx-autodoc
autodoc_default_options = {"ignore-module-all": True}
# Include __init__() docstring in class docstring
autoclass_content = "both"
autodoc_typehints = "description"
autodoc_typehints_description_target = "documented_params"
autodoc_typehints_format = "short"

# nbsphinx
nbsphinx_epilog = """
----

This page has been automatically generated by nbsphinx_ and can be run as a
Jupyter_ notebook available in the `=freqfit repository
<https://github.com/cenpa/=freqfit/tree/main/docs/source/notebooks>`_.

.. _nbsphinx: https://nbsphinx.readthedocs.io/
.. _Jupyter: https://jupyter.org/
"""



"""
Stores the PDF for Gaussian on uniform background, parametrized in number of decays. This model lacks any information on the efficiency and exposure.
"""
import numba as nb
import numpy as np

import freqfit.models.constants as constants
from freqfit.utils import inspectparameters

nb_kwd = {
    "nopython": True,
    "parallel": False,
    "nogil": True,
    "cache": True,
    "fastmath": True,
    "inline": "always",
}

QBB = constants.QBB
N_A = constants.NA
M_A = constants.MDET

# default analysis window and width
# window
#     uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
#     where edges of window are monotonically increasing (this is not checked), in keV.
#     Default is typical analysis window.
WINDOW = np.array(constants.WINDOW)

WINDOWSIZE = 0.0
for i in range(len(WINDOW)):
    WINDOWSIZE += WINDOW[i][1] - WINDOW[i][0]

SEED = 42  # set the default random seed


@nb.jit(**nb_kwd)
def nb_pdf(
    Es: np.array,
    n_S: float,
    n_B: float,
    delta: float,
    sigma: float,
    check_window: bool = False,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    n_S
        The signal rate, in units of counts
    n_B
        The background index rate, in counts
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    check_window
        whether to check if the passed Es fall inside of the window. Default is False and assumes that the passed Es
        all fall inside the window (for speed)

    Notes
    -----
    This function computes the following:
    mu_S = n_S
    mu_B = n_B
    pdf(E) = 1/(mu_S+mu_B) * [mu_S * norm(E_j, QBB - delta, sigma) + mu_B/windowsize]
    """

    # Precompute the signal and background counts
    mu_S = n_S
    mu_B = n_B

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = mu_B / WINDOWSIZE
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = (1 / (mu_S + mu_B)) * (
            S_amp * np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2)) + B_amp
        )

    if check_window:
        for i in nb.prange(Es.shape[0]):
            inwindow = False
            for j in range(len(WINDOW)):
                if WINDOW[j][0] <= Es[i] <= WINDOW[j][1]:
                    inwindow = True
            if not inwindow:
                y[i] = 0.0

    return y


@nb.jit(**nb_kwd)
def nb_density(
    Es: np.array,
    n_S: float,
    n_B: float,
    delta: float,
    sigma: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    n_S
        The signal rate, in units of counts
    n_B
        The background index rate, in counts
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV

    Notes
    -----
    This function computes the following, faster than without a numba wrapper:
    mu_S = n_S
    mu_B = n_B
    CDF(E) = mu_S + mu_B
    pdf(E) =[mu_S * norm(E_j, QBB - delta, sigma) + mu_B/windowsize]
    """

    mu_S = n_S
    mu_B = n_B

    if sigma == 0:
        return np.inf, np.full_like(Es, np.inf, dtype=np.float64)

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = mu_B / WINDOWSIZE
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = S_amp * np.exp(-((Es[i] - QBB + delta) ** 2) / (2 * sigma**2)) + B_amp

    return mu_S + mu_B, y


@nb.jit(**nb_kwd)
def nb_log_density(
    Es: np.array,
    n_S: float,
    n_B: float,
    delta: float,
    sigma: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    n_S
        The signal rate, in units of counts
    n_B
        The background index rate, in counts
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    Notes
    -----
    This function computes the following, faster than without a numba wrapper:
    mu_S = n_S
    mu_B = n_B
    CDF(E) = mu_S + mu_B
    pdf(E) =[mu_S * norm(E_j, QBB - delta, sigma) + mu_B/windowsize]
    """
    raise NotImplementedError("not yet implemented , sorry!")


@nb.jit(**nb_kwd)
def nb_density_gradient(
    Es: np.array,
    n_S: float,
    n_B: float,
    delta: float,
    sigma: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    n_S
        The signal rate, in units of counts
    n_B
        The background index rate, in counts
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV

    Notes
    -----
    This function computes the gradient of the density function and returns a tuple where the first element is the gradient of the CDF, and the second element is the gradient of the PDF.
    The first element has shape (K,) where K is the number of parameters, and the second element has shape (K,N) where N is the length of Es.
    mu_S = n_S
    mu_B = n_B
    pdf(E) = [mu_S * norm(E_j, QBB + delta, sigma) + mu_B/windowsize]
    cdf(E) = mu_S + mu_B
    """
    raise NotImplementedError("not yet implemented, sorry!")


@nb.jit(**nb_kwd)
def nb_logpdf(
    Es: np.array,
    n_S: float,
    n_B: float,
    delta: float,
    sigma: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    n_S
        The signal rate, in units of counts
    n_B
        The background index rate, in counts
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV

    Notes
    -----
    This function computes the following:
    mu_S = n_S
    mu_B = n_B
    logpdf(E) = log(1/(mu_S+mu_B) * [mu_S * norm(E_j, QBB - delta, sigma) + mu_B/windowsize])
    """
    raise NotImplementedError("not yet implemented!")


@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_rvs(
    n_sig: int,
    n_bkg: int,
    delta: float,
    sigma: float,
    seed: int = SEED,
) -> np.array:
    """
    Parameters
    ----------
    n_sig
        Number of signal events to pull from
    n_bkg
        Number of background events to pull from
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    seed
        specify a seed, otherwise uses default seed

    Notes
    -----
    This function pulls from a Gaussian for signal events and from a uniform distribution for background events
    in the provided windows, which may be discontinuous.
    """

    np.random.seed(seed)

    # Get energy of signal events from a Gaussian distribution
    # preallocate for background draws
    Es = np.append(np.random.normal(QBB - delta, sigma, size=n_sig), np.zeros(n_bkg))

    # Get background events from a uniform distribution
    bkg = np.random.uniform(0, 1, n_bkg)

    breaks = np.zeros(shape=(len(WINDOW), 2))
    for i in range(len(WINDOW)):
        thiswidth = WINDOW[i][1] - WINDOW[i][0]

        if i > 0:
            breaks[i][0] = breaks[i - 1][1]

        if i < len(WINDOW):
            breaks[i][1] = breaks[i][0] + thiswidth / WINDOWSIZE

        for j in range(len(bkg)):
            if breaks[i][0] <= bkg[j] <= breaks[i][1]:
                Es[n_sig + j] = (bkg[j] - breaks[i][0]) * thiswidth / (
                    breaks[i][1] - breaks[i][0]
                ) + WINDOW[i][0]

    return Es


@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_extendedrvs(
    n_S: float,
    n_B: float,
    delta: float,
    sigma: float,
    seed: int = SEED,
) -> np.array:
    """
    Parameters
    ----------
    n_S
        expected rate of signal events in events
    n_B
        rate of background events in events
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    seed
        specify a seed, otherwise uses default seed

    Notes
    -----
    This function pulls from a Gaussian for signal events and from a uniform distribution for background events
    in the provided windows, which may be discontinuous.
    """

    np.random.seed(seed)

    n_sig = n_S
    n_bkg = n_B

    # Get energy of signal events from a Gaussian distribution
    # preallocate for background draws
    Es = np.append(np.random.normal(QBB - delta, sigma, size=n_sig), np.zeros(n_bkg))

    # Get background events from a uniform distribution
    bkg = np.random.uniform(0, 1, n_bkg)

    breaks = np.zeros(shape=(len(WINDOW), 2))
    for i in range(len(WINDOW)):
        thiswidth = WINDOW[i][1] - WINDOW[i][0]

        if i > 0:
            breaks[i][0] = breaks[i - 1][1]

        if i < len(WINDOW):
            breaks[i][1] = breaks[i][0] + thiswidth / WINDOWSIZE

        for j in range(len(bkg)):
            if breaks[i][0] <= bkg[j] <= breaks[i][1]:
                Es[n_sig + j] = (bkg[j] - breaks[i][0]) * thiswidth / (
                    breaks[i][1] - breaks[i][0]
                ) + WINDOW[i][0]

    return Es, (n_sig, n_bkg)


class number_of_counts_0vbb_gen:
    def __init__(self):
        self.parameters = inspectparameters(self.density)
        pass

    def pdf(
        self,
        Es: np.array,
        n_S: float,
        n_B: float,
        delta: float,
        sigma: float,
        check_window: bool = False,
    ) -> np.array:
        return nb_pdf(Es, n_S, n_B, delta, sigma, check_window)

    def logpdf(
        self,
        Es: np.array,
        n_S: float,
        n_B: float,
        delta: float,
        sigma: float,
    ) -> np.array:
        return nb_logpdf(Es, n_S, n_B, delta, sigma)

    # for iminuit ExtendedUnbinnedNLL
    def density(
        self,
        Es: np.array,
        n_S: float,
        n_B: float,
        delta: float,
        sigma: float,
    ) -> np.array:
        return nb_density(Es, n_S, n_B, delta, sigma)

    # for iminuit ExtendedUnbinnedNLL
    def density_gradient(
        self,
        Es: np.array,
        n_S: float,
        n_B: float,
        delta: float,
        sigma: float,
    ) -> np.array:
        return nb_density_gradient(Es, n_S, n_B, delta, sigma)

    # for iminuit ExtendedUnbinnedNLL
    def log_density(
        self,
        Es: np.array,
        n_S: float,
        n_B: float,
        delta: float,
        sigma: float,
    ) -> np.array:
        return nb_log_density(Es, n_S, n_B, delta, sigma)

    # should we have an rvs method for drawing a random number of events?
    # `extendedrvs`
    # needs to use same parameters as the rest of the functions...
    def rvs(
        self,
        n_sig: int,
        n_bkg: int,
        delta: float,
        sigma: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_rvs(n_sig, n_bkg, delta, sigma, seed=seed)

    def extendedrvs(
        self,
        n_S: float,
        n_B: float,
        delta: float,
        sigma: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_extendedrvs(n_S, n_B, delta, sigma, seed=seed)

    def plot(
        self,
        Es: np.array,
        n_S: float,
        n_B: float,
        delta: float,
        sigma: float,
    ) -> None:
        y = nb_pdf(Es, n_S, n_B, delta, sigma)

        import matplotlib.pyplot as plt

        plt.step(Es, y)
        plt.show()

    # function call needs to take the same parameters as the other function calls, in the same order repeated twice
    # this is intended only for empty datasets
    # returns `None` if we couldn't combine the datasets (a dataset was not empty)
    def combine(
        self,
        a_Es: np.array,
        a_n_S: float,
        a_n_B: float,
        a_delta: float,
        a_sigma: float,
        b_Es: np.array,
        b_n_S: float,
        b_n_B: float,
        b_delta: float,
        b_sigma: float,
    ) -> list | None:
        # datasets must be empty to be combined
        if len(a_Es) != 0 or len(b_Es) != 0:
            return None

        Es = np.array([])  # both of these datasets are empty
        n_S = 0.0  # this should be overwritten in the fit later
        # n_B = 0.0  # this should be overwritten in the fit later

        # exp = 1 + 1  # total exposure

        # # exposure weighted fixed parameters (important to calculate correctly)
        # sigma = (a_sigma + b_exp * b_sigma) / exp
        # eff = (a_exp * a_eff + b_exp * b_eff) / exp
        # delta = (a_exp * a_delta + b_exp * b_delta) / exp

        # # these are fully correlated in this model so the direct sum is appropriate
        # # (maybe still appropriate even if not fully correlated?)
        # effunc = (a_exp * a_effunc + b_exp * b_effunc) / exp

        # effuncscale = 0.0  # this should be overwritten in the fit later

        # return [Es, S, BI, delta, sigma, eff, effunc, effuncscale, exp]
        return None

    def can_combine(
        self,
        a_Es: np.array,
        a_n_S: float,
        a_n_B: float,
        a_delta: float,
        a_sigma: float,
    ) -> bool:
        """
        This sets an arbitrary rule if this dataset can be combined with other datasets.
        In this case, if the dataset contains no data, then it can be combined, but more complex rules can be imposed.
        """
        if len(a_Es) == 0:
            return True
        else:
            return False


number_of_counts_0vbb = number_of_counts_0vbb_gen()



"""
This sphinx extension aims to improve the documentation of numba-decorated
functions. It it inspired by the design of Celery's sphinx extension
'celery.contrib.sphinx'.

Adapted from https://github.com/numba/numba/issues/5755#issuecomment-646587651
"""

from copy import copy
from typing import Iterator, List

from docutils import nodes
from numba.core.dispatcher import Dispatcher
from numba.np.ufunc.dufunc import DUFunc
from numba.np.ufunc.gufunc import GUFunc
from sphinx.domains.python import PyFunction
from sphinx.ext.autodoc import FunctionDocumenter


class NumbaFunctionDocumenter(FunctionDocumenter):
    """Document numba decorated functions."""

    def import_object(self) -> bool:
        """Import the object given by *self.modname* and *self.objpath* and set
        it as *self.object*.

        Returns True if successful, False if an error occurred.
        """
        success = super().import_object()
        if success:
            # Store away numba wrapper
            self.jitobj = self.object
            # And bend references to underlying python function
            if hasattr(self.object, "py_func"):
                self.object = self.object.py_func
            elif hasattr(self.object, "_dispatcher") and hasattr(
                self.object._dispatcher, "py_func"
            ):
                self.object = self.object._dispatcher.py_func
            elif hasattr(self.object, "gufunc_builder") and hasattr(
                self.object.gufunc_builder, "py_func"
            ):
                self.object = self.object.gufunc_builder.py_func
            else:
                success = False
        return success

    def process_doc(self, docstrings: List[List[str]]) -> Iterator[str]:
        """Let the user process the docstrings before adding them."""
        # Essentially copied from FunctionDocumenter
        for docstringlines in docstrings:
            if self.env.app:
                # let extensions preprocess docstrings
                # need to manually set 'what' to FunctionDocumenter.objtype
                # to not confuse preprocessors like napoleon with an objtype
                # that they don't know
                self.env.app.emit(
                    "autodoc-process-docstring",
                    FunctionDocumenter.objtype,
                    self.fullname,
                    self.object,
                    self.options,
                    docstringlines,
                )

            extra_lines = []

            # This block inserts information about precompiled signatures
            if getattr(self.jitobj, "types", []) and getattr(
                self.jitobj, "_frozen", False
            ):
                line = "``" + "``, ``".join(self.jitobj.types) + "``"
                extra_lines.insert(0, "- *Precompiled signatures:* " + line)

            if hasattr(self.jitobj, "gufunc_builder"):
                if hasattr(self.jitobj.gufunc_builder, "targetoptions"):
                    opts = copy(self.jitobj.gufunc_builder.targetoptions)
                else:
                    opts = {}

                if hasattr(self.jitobj.gufunc_builder, "cache"):
                    opts["cache"] = self.jitobj.gufunc_builder.cache

                line = []
                for k, v in opts.items():
                    line.append(f"{k}={v}")

                if line:
                    line = "``" + "``, ``".join(line) + "``"
                    extra_lines.insert(0, "- *Options:* " + line)

            if extra_lines:
                docstringlines = extra_lines + [""] + docstringlines

            yield from docstringlines


class JitDocumenter(NumbaFunctionDocumenter):
    """Document jit/njit decorated functions."""

    objtype = "jitfun"

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        return isinstance(member, Dispatcher) and hasattr(member, "py_func")


class VectorizeDocumenter(NumbaFunctionDocumenter):
    """Document vectorize decorated functions."""

    objtype = "vecfun"

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        return (
            isinstance(member, DUFunc)
            and hasattr(member, "_dispatcher")
            and hasattr(member._dispatcher, "py_func")
        )


class GUVectorizeDocumenter(NumbaFunctionDocumenter):
    """Document guvectorize decorated functions."""

    objtype = "guvecfun"

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        return (
            isinstance(member, GUFunc)
            and hasattr(member, "gufunc_builder")
            and hasattr(member.gufunc_builder, "py_func")
        )


class JitDirective(PyFunction):
    """Sphinx jitfun directive."""

    def get_signature_prefix(self, sig):
        return [nodes.Text("@numba.jit ")]


class VectorizeDirective(PyFunction):
    """Sphinx vecfun directive."""

    def get_signature_prefix(self, sig):
        return [nodes.Text("@numba.vectorize ")]


class GUVectorizeDirective(PyFunction):
    """Sphinx guvecfun directive."""

    def get_signature_prefix(self, sig):
        return [nodes.Text("@numba.guvectorize ")]


def setup(app):
    """Setup Sphinx extension."""
    # Register the new documenters and directives (autojitfun, autovecfun)
    # Set the default prefix which is printed in front of the function signature
    app.setup_extension("sphinx.ext.autodoc")
    app.add_autodocumenter(JitDocumenter)
    app.add_directive_to_domain("py", "jitfun", JitDirective)
    app.add_autodocumenter(VectorizeDocumenter)
    app.add_directive_to_domain("py", "vecfun", VectorizeDirective)
    app.add_autodocumenter(GUVectorizeDocumenter)
    app.add_directive_to_domain("py", "guvecfun", GUVectorizeDirective)

    return {"parallel_read_safe": True}



import numpy as np
from scipy.stats import chi2

from freqfit import Experiment
from freqfit.statistics import dkw_band, emp_cdf, ts_critical


def test_dkw():
    data = [
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        3,
        3,
        3,
        3,
        3,
        4,
        4,
        4,
        4,
        4,
        5,
        5,
        5,
        5,
        5,
        6,
        6,
        6,
        6,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
    ]
    bins = [
        0.5,
        1.5,
        2.5,
        3.5,
        4.5,
        5.5,
        6.5,
        7.5,
        8.5,
        9.5,
        10.5,
        11.5,
        12.5,
        13.5,
        14.5,
        15.5,
        16.5,
    ]

    cdf, _ = emp_cdf(data=data, bins=bins)

    dkw_lo, dkw_hi = dkw_band(cdf, len(data))

    hand_cdf = np.array(
        [
            0.2,
            0.4,
            0.5,
            0.6,
            0.7,
            0.8,
            0.82,
            0.84,
            0.86,
            0.88,
            0.9,
            0.92,
            0.94,
            0.96,
            0.98,
            1,
        ]
    )
    hand_dkw_hi = np.minimum(hand_cdf + 0.13537, 1)
    hand_dkw_lo = np.maximum(hand_cdf - 0.13537, 0)

    assert (hand_cdf == cdf).all()
    assert np.isclose(dkw_lo, hand_dkw_lo, atol=1e-5).all()
    assert np.isclose(dkw_hi, hand_dkw_hi, atol=1e-5).all()


def test_criticalts():
    true_S = 10.0
    p = Experiment.file("tests/config_test_highsignal.yaml", "experiment")

    # profile over the test statistic
    x = np.arange(0, 30, 0.1)
    y = np.zeros_like(x)
    for i, xx in enumerate(x):
        y[i], *_ = p.ts({"global_S": xx})

    toypars = p.profile({"global_S": true_S})["values"]
    toypars["global_S"] = true_S

    numtoys = 2000
    toyts = p.toy_ts(toypars, {"global_S": true_S}, num=numtoys)[0]
    nbins = 500

    (crit95, lo95, hi95), _ = ts_critical(
        toyts, bins=nbins, confidence=0.99, threshold=0.95
    )
    assert lo95 < chi2.ppf(0.95, df=1) and chi2.ppf(0.95, df=1) < hi95

    (crit90, lo90, hi90), _ = ts_critical(
        toyts, bins=nbins, confidence=0.99, threshold=0.9
    )
    assert lo90 < chi2.ppf(0.9, df=1) and chi2.ppf(0.9, df=1) < hi90

    (crit68, lo68, hi68), _ = ts_critical(
        toyts, bins=nbins, confidence=0.99, threshold=0.68
    )
    assert lo68 < chi2.ppf(0.68, df=1) and chi2.ppf(0.68, df=1) < hi68



"""
A class that controls an experiment and calls the `Superset` class.
"""
import itertools
import logging
from copy import deepcopy

import numpy as np
from iminuit import Minuit

from freqfit import initial_guesses
from freqfit.superset import Superset
from freqfit.toy import Toy
from freqfit.utils import grab_results, load_config

SEED = 42

log = logging.getLogger(__name__)


class Experiment(Superset):
    def __init__(
        self,
        config: dict,
        name: str = "",
    ) -> None:
        self.toy = None  # the last Toy from this experiment
        self.best = None  # to store the best fit result
        self.guess = None  # store the initial guess
        self.minuit = None  # Minuit object

        self.try_to_combine_datasets = False
        self.test_statistic = "t_mu"

        # minimization options
        self.backend = "minuit"  # "minuit" or "scipy"
        self.scipy_minimizer = None
        self.use_log = False  # Option in the config to use the logdensity instead of the density in the cost function
        self.iminuit_tolerance = 1e-5  # tolerance for iminuit or other minimizer
        self.iminuit_precision = 1e-10  # precision for the Iminuit minimizer, especially important for scipy minimizers
        self.iminuit_strategy = 0
        self.minimizer_options = {}  # dict of options to pass to the iminuit minimizer
        self.use_grid_rounding = False  # whether to stick parameters on a grid when evaluating the test statistic after minimizing
        self.grid_rounding_num_decimals = (
            {}
        )  # dict telling us how many decimals to round the parameters to when grid rounding

        self.scan = False
        self.scan_grid = None  # this is a dictionary, each key is a fit parameter and its values are the parameter values to scan along in one dimension
        self.user_gradient = (
            False  # option to use a user-specified density gradient for a model
        )
        self.data = (
            []
        )  # A flat array of all the data. The data may be split between datasets, this is just an aggregate
        self.initial_guess_function = None

        self.fixed_bc_no_data = (
            {}
        )  # check which nuisance parameters can be fixed in the fit due to no data

        combined_datasets = None
        if "options" in config:
            if "try_to_combine_datasets" in config["options"]:
                self.try_to_combine_datasets = config["options"][
                    "try_to_combine_datasets"
                ]
                if "combined_datasets" in config:
                    combined_datasets = config["combined_datasets"]
                    msg = "found 'combined_datasets' in config"
                    logging.info(msg)

            if "name" in config["options"] and (name == ""):
                name = config["options"]["name"]

            if "backend" in config["options"]:
                self.backend = config["options"]["backend"]

            if "iminuit_tolerance" in config["options"]:
                self.iminuit_tolerance = config["options"]["iminuit_tolerance"]

            if "iminuit_precision" in config["options"]:
                self.iminuit_precision = config["options"]["iminuit_precision"]

            if "minimizer_options" in config["options"]:
                self.minimizer_options = config["options"]["minimizer_options"]

                if not isinstance(self.minimizer_options, dict):
                    raise ValueError("minimizer_options must be a dictionary")

            if self.backend == "scipy":
                if "scipy_minimizer" in config["options"]:
                    self.scipy_minimizer = config["options"]["scipy_minimizer"]
                    if self.scipy_minimizer not in [
                        "Nelder-Mead",
                        "Powell",
                        "CG",
                        "BFGS",
                        "Newton-CG",
                        "L-BFGS-B",
                        "TNC",
                        "COBYLA",
                        "COBYQA",
                        "SLSQP",
                        "trust-constr",
                        "dogleg",
                        "trust-ncg",
                        "trust-exact",
                        "trust-krylov",
                        None,
                    ]:
                        raise NotImplementedError(
                            f"{self.scipy_minimizer} is not a valid minimizer"
                        )

            if "user_gradient" in config["options"]:
                self.user_gradient = config["options"]["user_gradient"]

            if "use_log" in config["options"]:
                self.use_log = config["options"]["use_log"]

            if "scan" in config["options"]:
                self.scan = config["options"]["scan"]

                if "scan_grid" in config["options"]:
                    self.scan_grid = config["options"]["scan_grid"]

            if "initial_guess_function" in config["options"]:
                self.initial_guess_function = config["options"][
                    "initial_guess_function"
                ]

            if "use_grid_rounding" in config["options"]:
                self.use_grid_rounding = config["options"]["use_grid_rounding"]
                logging.info("using grid rounding")

            if "test_statistic" in config["options"]:
                if config["options"]["test_statistic"] in [
                    "t_mu",
                    "t_mu_tilde",
                    "q_mu",
                    "q_mu_tilde",
                ]:
                    self.test_statistic = config["options"]["test_statistic"]
                    msg = f"setting test statistic: {self.test_statistic}"
                    logging.info(msg)
            else:
                msg = "setting test statistic to default: t_mu"
                logging.info(msg)

        constraints = None
        if "constraints" in config:
            constraints = config["constraints"]
            msg = "found 'constraints' in config"
            logging.info(msg)
        else:
            msg = "did not find 'constraints' in config"
            logging.info(msg)

        super().__init__(
            datasets=config["datasets"],
            parameters=config["parameters"],
            constraints=constraints,
            combined_datasets=combined_datasets,
            name=name,
            try_to_combine_datasets=self.try_to_combine_datasets,
            user_gradient=self.user_gradient,
            use_log=self.use_log,
        )

        # get the fit parameters and set the parameter initial values
        self.guess = self.initialguess()
        self.minuit = Minuit(self.costfunction, **self.guess)
        self.minuit.tol = self.iminuit_tolerance  # set the tolerance
        self.minuit.precision = self.iminuit_precision
        self.minuit.strategy = self.iminuit_strategy

        # raise a RunTime error if function evaluates to NaN
        self.minuit.throw_nan = True

        # check which nuisance parameters can be fixed in the fit due to no data

        # find which parameters are part of Datasets that have data
        parstofitthathavedata = set()
        for datasetname in self.datasets:
            # check if there is some data
            if self.datasets[datasetname].data.size > 0:
                # Add the data to self.data
                self.data.extend(self.datasets[datasetname].data)
                # add the fit parameters of this Dataset if there is some data
                for fitpar in self.datasets[datasetname].fitparameters:
                    parstofitthathavedata.add(fitpar)

        # now check which fit parameters can be fixed if no data and are not part of a Dataset that has data
        for parname in self.fitparameters:
            if (
                "fix_if_no_data" in self.parameters[parname]
                and self.parameters[parname]["fix_if_no_data"]
            ):
                # check if this parameter is part of a Dataset that has data
                if parname not in parstofitthathavedata:
                    self.fixed_bc_no_data[parname] = self.parameters[parname]["value"]

        # set the grid rounding
        if self.use_grid_rounding:
            for parname in list(config["parameters"].keys()):
                if "grid_rounding_num_decimals" in self.parameters[parname]:
                    self.grid_rounding_num_decimals[parname] = self.parameters[parname][
                        "grid_rounding_num_decimals"
                    ]
                else:
                    self.grid_rounding_num_decimals[parname] = 128  # something big

        # to set limits and fixed variables
        # this function will also fix those fit parameters which can be fixed because they are not part of a
        # Dataset that has data
        self.minuit_reset()

        # If scan is not none, initialize the hypercube grid
        if (self.scan) and (self.scan_grid is not None):
            self.hypercube_grid = self.create_hypercube(self.scan_grid)

    @classmethod
    def file(
        cls,
        file: str,
        name: str = "",
    ):
        logging.warning('Experiment.file() will be deprecated, use Experiment.from_file() instead')
        config = load_config(file=file)
        return cls(config=config, name=name)

    @classmethod
    def from_file(
        cls,
        file: str,
        name: str = "",
    ):
        config = load_config(file=file)
        return cls(config=config, name=name)

    @classmethod
    def from_dict(
        cls,
        input: str,
        name: str = "",
    ):
        config = load_config(file=input)
        return cls(config=config, name=name)
        
    def initialguess(
        self,
    ) -> dict:
        if self.initial_guess_function is None:
            guess = {
                fitpar: self.parameters[fitpar]["value"]
                if "value" in self.parameters[fitpar]
                else None
                for fitpar in self.fitparameters
            }

        else:
            func = getattr(initial_guesses, self.initial_guess_function)
            guess = func(self)

        # for fitpar, value in guess.items():
        #     if value is None:
        #         guess[fitpar] = 1e-9

        # could put other stuff here to get a better initial guess though probably that should be done
        # somewhere specific to the analysis. Or maybe this function could call something for the analysis.
        # eh, probably better to make the initial guess elsewhere and stick it in the config file.
        # for that reason, I'm commenting out the few lines above.

        return guess

    def minuit_reset(
        self,
        use_physical_limits: bool = True,  # for numerators of test statistics, want this to be False
    ) -> None:
        # resets the minimization and stuff
        # does not change limits but does remove "fixed" attribute of variables
        # 2024/08/09: This no longer seems to be true? Not sure if something changed in iminuit or if I was wrong?
        self.minuit.reset()

        # Restore the first initial guess
        self.minuit.values[:] = list(self.guess.values())

        # overwrite the limits
        # note that this information can also be contained in a Dataset when instantiated
        # and is overwritten here

        # also set which parameters are fixed
        for parname, pardict in self.parameters.items():
            if parname in self.minuit.parameters:
                self.minuit.fixed[parname] = False
                self.minuit.limits[parname] = (-1.0 * np.inf, np.inf)

                if "limits" in pardict:
                    self.minuit.limits[parname] = pardict["limits"]
                if use_physical_limits and "physical_limits" in pardict:
                    self.minuit.limits[parname] = pardict["physical_limits"]
                if "fixed" in pardict:
                    self.minuit.fixed[parname] = pardict["fixed"]
                # fix those nuisance parameters which can be fixed because they are not part of a
                # Dataset that has data
                if parname in self.fixed_bc_no_data:
                    self.minuit.fixed[parname] = True
                    self.minuit.values[parname] = self.fixed_bc_no_data[parname]

        return

    def bestfit(
        self,
        force: bool = False,
        use_physical_limits: bool = True,
    ) -> dict:
        """
        force
            By default (`False`), if `self.best` has a result, the minimization will be skipped and that
            result will be returned instead. If `True`, the minimization will be run and the result returned and
            stored as `self.best`

        Performs a global minimization and returns a `dict` with the results. These results are also stored in
        `self.best`.
        """
        # don't run this more than once if we don't have to
        if self.best is not None and not force:
            return self.best

        # remove any previous minimizations
        self.minuit_reset(use_physical_limits=use_physical_limits)

        if self.scan:
            y = np.empty(len(self.hypercube_grid))
            for i in range(len(self.hypercube_grid)):
                y[i] = self.minuit._fcn(self.hypercube_grid[i])

            best = np.min(y)
            ibest = np.argmin(y)

            self.best = {
                "fval": best,
                "values": {},
                "valid": True,
            }  # TODO: add the rest of return values of interest, like parameter names at minimum
            for par in self.minuit.parameters:
                ipar, vpar = self.minuit._normalize_key(par)
                self.best["values"][par] = self.hypercube_grid[ibest][ipar]

        else:
            try:
                if self.backend == "minuit":
                    self.minuit.migrad(**self.minimizer_options)
                elif self.backend == "scipy":
                    self.minuit.scipy(
                        method=self.scipy_minimizer, options=self.minimizer_options
                    )
                else:
                    raise NotImplementedError(
                        "Iminuit backend is not set to `minuit` or `scipy`"
                    )
            except RuntimeError:
                msg = "`Experiment` has invalid best fit"
                logging.debug(msg)

            if not self.minuit.valid:
                msg = "`Experiment` has invalid best fit"
                logging.debug(msg)

            self.best = grab_results(
                self.minuit,
                use_grid_rounding=self.use_grid_rounding,
                grid_rounding_num_decimals=self.grid_rounding_num_decimals,
            )

        if "global_S" in self.best["values"]:
            if self.best["values"]["global_S"] < 1e-20:
                self.best = self.profile({"global_S": 0.0})

        if self.guess == self.best["values"]:
            msg = "`Experiment` has best fit values very close to initial guess"
            logging.debug(msg)

        return self.best

    def profile(
        self,
        parameters: dict,
        use_physical_limits: bool = True,
    ) -> dict:
        """
        parameters
            `dict` where keys are names of parameters to fix and values are the value that the parameter should be
            fixed to
        """

        self.minuit_reset(use_physical_limits=use_physical_limits)

        # set fixed parameters
        for parname, parvalue in parameters.items():
            self.minuit.fixed[parname] = True
            self.minuit.values[parname] = parvalue

        if self.scan:
            # need to remake the hypercube because there are now fewer parameters
            # pop the fixed parameters from the supplied scan_grid
            new_grid_dict = deepcopy(self.scan_grid)
            for parname in parameters.keys():
                if parname in new_grid_dict.keys():
                    new_grid_dict.pop(parname)
            hypercube_grid = self.create_hypercube(new_grid_dict)

            y = np.empty(len(hypercube_grid))
            for i in range(len(hypercube_grid)):
                y[i] = self.minuit._fcn(hypercube_grid[i])

            best = np.min(y)
            ibest = np.argmin(y)
            results = {
                "fval": best,
                "values": {},
                "valid": True,
            }  # TODO: add the rest of return values of interest
            for par in self.minuit.parameters:
                ipar, vpar = self.minuit._normalize_key(par)
                results["values"][par] = hypercube_grid[ibest][ipar]

        else:
            try:
                if self.backend == "minuit":
                    self.minuit.migrad(**self.minimizer_options)
                elif self.backend == "scipy":
                    self.minuit.scipy(
                        method=self.scipy_minimizer, options=self.minimizer_options
                    )
                else:
                    raise NotImplementedError(
                        "Iminuit backend is not set to `minuit` or `scipy`"
                    )
            except RuntimeError:
                msg = f"`Experiment` throwing NaN has invalid profile at {parameters}"
                logging.debug(msg)

            if not self.minuit.valid:
                msg = "`Experiment` has invalid profile"
                logging.debug(msg)

            results = grab_results(
                self.minuit,
                use_grid_rounding=self.use_grid_rounding,
                grid_rounding_num_decimals=self.grid_rounding_num_decimals,
            )

        if self.guess == results["values"]:
            msg = "`Experiment` has profile fit values very close to initial guess"
            logging.debug(msg)

        # also include the fixed parameters
        for parname, parvalue in parameters.items():
            results["values"][parname] = parvalue

        return results

    # this corresponds to t_mu or t_mu^tilde depending on whether there is a physical limit on the parameters
    def ts(
        self,
        profile_parameters: dict,  # which parameters to fix and their value (rest are profiled)
        force: bool = False,
    ) -> float:
        """
        force
            See `experiment.bestfit()` for description. Default is `False`.
        """

        use_physical_limits = False  # for t_mu and q_mu
        if self.test_statistic == "t_mu_tilde" or self.test_statistic == "q_mu_tilde":
            use_physical_limits = True

        denom = self.bestfit(force=force, use_physical_limits=use_physical_limits)[
            "fval"
        ]

        # see Cowan (2011) Eq. 14 and Eq. 16
        if self.test_statistic == "q_mu" or self.test_statistic == "q_mu_tilde":
            for parname, parvalue in profile_parameters.items():
                if self.best["values"][parname] > parvalue:
                    return 0.0, 0.0, 0.0

        num = self.profile(
            parameters=profile_parameters, use_physical_limits=use_physical_limits
        )["fval"]

        ts = num - denom

        if ts < 0:
            msg = f"`Experiment` gave test statistic below zero: {ts}"
            logging.debug(msg)

        # because these are already -2*ln(L) from iminuit
        return ts, denom, num

    def maketoy(
        self,
        parameters: dict,
        seed: int = SEED,
    ) -> Toy:
        self.toy = Toy(experiment=self, parameters=parameters, seed=seed)

        return self.toy

    def toy_ts(
        self,
        parameters: dict,  # parameters and values needed to generate the toys
        profile_parameters: dict
        | list,  # which parameters to fix and their value (rest are profiled)
        num: int = 1,
        seed: np.array = None,
    ):
        """
        Makes a number of toys and returns their test statistics.
        Having the seed be an array allows for different jobs producing toys on the same s-value to have different seed numbers
        """
        np.random.seed(SEED)
        if seed is None:
            seed = np.random.randint(1000000, size=num)
        else:
            if len(seed) != num:
                raise ValueError("Seeds must have same length as the number of toys!")
        if isinstance(profile_parameters, dict):
            ts = np.zeros(num)
            numerators = np.zeros(num)
            denominators = np.zeros(num)
            data_to_return = []
            paramvalues_to_return = []
            num_drawn = []
            for i in range(num):
                thistoy = self.maketoy(parameters=parameters, seed=seed[i])
                ts[i], denominators[i], numerators[i] = thistoy.ts(
                    profile_parameters=profile_parameters
                )
                data_to_return.append(thistoy.toy_data_to_save)
                paramvalues_to_return.append(thistoy.parameters_to_save)
                num_drawn.append(thistoy.toy_num_drawn_to_save)

        # otherwise profile parameters is a list of dicts
        else:
            ts = np.zeros((len(profile_parameters), num))
            numerators = np.zeros((len(profile_parameters), num))
            denominators = np.zeros((len(profile_parameters), num))
            data_to_return = []
            paramvalues_to_return = []
            num_drawn = []
            for i in range(num):
                thistoy = self.maketoy(parameters=parameters, seed=seed[i])
                for j in range(len(profile_parameters)):
                    ts[j][i], denominators[j][i], numerators[j][i] = thistoy.ts(
                        profile_parameters=profile_parameters[j]
                    )
                data_to_return.append(thistoy.toy_data_to_save)
                paramvalues_to_return.append(thistoy.parameters_to_save)
                num_drawn.append(thistoy.toy_num_drawn_to_save)

        # Need to flatten the data_to_return in order to save it in h5py
        data_to_return_flat = (
            np.ones(
                (len(data_to_return), np.nanmax([len(arr) for arr in data_to_return]))
            )
            * np.nan
        )
        for i, arr in enumerate(data_to_return):
            data_to_return_flat[i, : len(arr)] = arr

        num_drawn_to_return_flat = (
            np.ones((len(num_drawn), np.nanmax([len(arr) for arr in num_drawn])))
            * np.nan
        )
        for i, arr in enumerate(num_drawn):
            num_drawn_to_return_flat[i, : len(arr)] = arr

        return (
            ts,
            data_to_return_flat,
            paramvalues_to_return,
            num_drawn_to_return_flat,
            denominators,
            numerators,
        )

    # mostly pulled directly from iminuit, with some modifications to ignore empty Datasets and also to format
    # plots slightly differently
    def visualize(
        self,
        parameters: dict,
        component_kwargs=None,
    ) -> None:
        """
        Visualize data and model agreement (requires matplotlib).

        The visualization is drawn with matplotlib.pyplot into the current figure.
        Subplots are created to visualize each part of the cost function, the figure
        height is increased accordingly. Parts without a visualize method are silently
        ignored.

        Does not draw Datasets that are empty (have no events).

        Parameters
        ----------
        args : array-like
            Parameter values.
        component_kwargs : dict of dicts, optional
            Dict that maps an index to dict of keyword arguments. This can be
            used to pass keyword arguments to a visualize method of a component with
            that index.
        **kwargs :
            Other keyword arguments are forwarded to all components.
        """
        from matplotlib import pyplot as plt

        args = []
        for par in self.fitparameters:
            if par not in parameters:
                msg = f"parameter {par} was not provided"
                raise KeyError(msg)
            args.append(parameters[par])

        n = 0
        for comp in self.costfunction:
            if hasattr(comp, "visualize"):
                if hasattr(comp, "data"):
                    if len(comp.data) > 0:
                        n += 1
                else:
                    n += 1

        fig = plt.gcf()
        fig.set_figwidth(n * fig.get_figwidth() / 1.5)
        _, ax = plt.subplots(1, n, num=fig.number)

        if component_kwargs is None:
            component_kwargs = {}

        i = 0
        for k, (comp, cargs) in enumerate(self.costfunction._split(args)):
            if not hasattr(comp, "visualize"):
                continue
            if hasattr(comp, "data") and len(comp.data) == 0:
                continue
            kwargs = component_kwargs.get(k, {})
            plt.sca(ax[i])
            comp.visualize(cargs, **kwargs)
            i += 1
        
        return fig

    # mostly pulled directly from iminuit, with some modifications to ignore empty Datasets and also to format
    # plots slightly differently
    def pull_plot(
        self,
        parameters: dict,
        component_kwargs=None,
    ) -> None:
        """
        Visualize data and model agreement (requires matplotlib).

        The visualization is drawn with matplotlib.pyplot into the current figure.
        Subplots are created to visualize each part of the cost function, the figure
        height is increased accordingly. Parts without a visualize method are silently
        ignored.

        Does not draw Datasets that are empty (have no events).

        Parameters
        ----------
        args : array-like
            Parameter values.
        component_kwargs : dict of dicts, optional
            Dict that maps an index to dict of keyword arguments. This can be
            used to pass keyword arguments to a visualize method of a component with
            that index.
        **kwargs :
            Other keyword arguments are forwarded to all components.
        """
        from matplotlib import pyplot as plt

        args = []
        for par in self.fitparameters:
            if par not in parameters:
                msg = f"parameter {par} was not provided"
                raise KeyError(msg)
            args.append(parameters[par])

        n = 0
        totnum = 0
        for comp in self.costfunction:
            totnum += 1
            if hasattr(comp, "visualize"):
                if hasattr(comp, "data"):
                    if len(comp.data) > 0:
                        n += 1
                else:
                    n += 1

        fig = plt.gcf()
        # fig.set_figwidth(n * fig.get_figwidth() / 1.5)
        _, ax = plt.subplots(1, 1, num=fig.number)

        if component_kwargs is None:
            component_kwargs = {}

        i = 0
        for k, (comp, cargs) in enumerate(self.costfunction._split(args)):
            # assumes nuisance constraints are the last cost function...
            if k != totnum - 1:
                continue
            if not hasattr(comp, "visualize"):
                continue
            if hasattr(comp, "data") and len(comp.data) == 0:
                continue
            kwargs = component_kwargs.get(k, {})
            plt.sca(ax)
            comp.visualize(cargs, **kwargs)
            i += 1
        
        return fig
        

    def create_hypercube(self, scan_grid) -> list:
        """
        Parameters
        ----------
        m
            Minuit object that holds that parameters and parameter names
        scan_grid
            The dictionary object that holds the points in each dimension to scan
        """
        scan_pars = list(scan_grid.keys())
        scan_list_of_lists = [list(scan_grid[scan_par]) for scan_par in scan_pars]
        hypercube_missing_other_values = itertools.product(
            *scan_list_of_lists
        )  # This list is missing all of the other function values that the minuit object needs, let's add them back in!

        # get the indices within the minuit object of all of parameters on the grid
        ipar_list = []
        for scan_par in scan_pars:
            ipar_list.append(self.minuit._normalize_key(scan_par)[0])
        ipar_list = np.array(ipar_list)
        values = deepcopy(
            np.array(self.minuit.values)
        )  # Get all the values, don't mutate them

        hypercube = []
        for par in hypercube_missing_other_values:
            values[ipar_list] = par
            hypercube.append(deepcopy(values))

        return hypercube



import numpy as np
from iminuit import cost

from freqfit.dataset import Dataset
from freqfit.models import gaussian_on_uniform


def test_dataset():
    data = np.array([1, 2, 3, 4])

    model = gaussian_on_uniform

    model_parameters = {
        "S": "global_S",
        "BI": "global_BI",
        "delta": "delta_dset0",
        "sigma": "sigma_dset0",
        "eff": "eff_dset0",
        "exp": "exp_dset0",
    }

    parameters = {
        "global_S": {"includeinfit": True, "value": 1e-9, "limits": [0, None]},
        "global_BI": {"includeinfit": True, "value": 1e-3, "limits": [0, None]},
        "delta_dset0": {"includeinfit": False, "value": 1e-1, "limits": None},
        "sigma_dset0": {"includeinfit": False, "value": 1e-2, "limits": None},
        "eff_dset0": {"includeinfit": False, "value": 1e-3, "limits": [0, 1]},
        "exp_dset0": {"includeinfit": False, "value": 1e-4, "limits": [0, None]},
    }

    costfunction = cost.ExtendedUnbinnedNLL
    name = "test_dset"
    dset = Dataset(data, model, model_parameters, parameters, costfunction, name)

    assert np.array_equal(dset.data, data)
    assert dset.name == name
    assert dset.model == model
    assert dset._parlist_indices == [
        0,
        1,
    ]  # We are fitting both S and BI, which are the first 2 parameters of the gaussian_on_uniform density
    assert np.array_equal(
        dset._parlist[:], [1e-9, 1e-3, 1e-1, 1e-2, 1e-3, 1e-4]
    )  # Check that the initial values are set correctly, exclude the window

    # Test the density function handling
    S = 1
    BI = 2
    reference_density = gaussian_on_uniform.density(data, S, BI, 1e-1, 1e-2, 1e-3, 1e-4)
    test_density = dset.density(data, S, BI)

    assert np.array_equal(test_density[0], reference_density[0])
    assert np.array_equal(test_density[1], reference_density[1])

    reference_density_gradient = gaussian_on_uniform.density_gradient(
        data, S, BI, 1e-1, 1e-2, 1e-3, 1e-4
    )
    test_density_gradient = dset.density_gradient(data, S, BI)

    # Test the density gradient function handling
    assert (
        len(test_density_gradient[0]) == 2
    )  # make sure we are masking things correctly
    mask_idxs = np.array([0, 1])
    assert np.array_equal(
        test_density_gradient[0], reference_density_gradient[0][mask_idxs]
    )  # only the first two parameters are being varied

    assert np.array_equal(
        test_density_gradient[1][0], reference_density_gradient[1][0, :]
    )  # only the first two parameters are being varied
    assert np.array_equal(
        test_density_gradient[1][1], reference_density_gradient[1][1, :]
    )  # only the first two parameters are being varied



"""A module for computing Feldman-Cousins Poisson confidence intervals.

Computes unified confidence intervals for Poisson counting experiments with a
known background level based on the Feldman-Cousins ordering principle:
Feldman and Cousins, Phys. Rev. D 57, 3873 (1998).

Uses analytical solutions and inversions to perform all computations on-the-fly,
so no memory allocation or setup is required: just call the desired function
with the appropriate parameters.

The results agree to 3 decimal places with nearly all entries in Table IV of the
Feldman-Cousins paper. Deviations were only observed for very high background
and very low observed counts. Investigations by the author revealed these
calculations to likely be more accurate than the published versions.

Similar / related projects:
- https://docs.gammapy.org/dev/stats/feldman_cousins.html
Requires lengthy precalculation of coverage maps based on user-supplied tables,
but handles other PDFs besides the Poisson distribution.
- https://github.com/mxmeier/feldman_cousins
Similar in functionality and usage to the gammapy package.

Example:

n = 3
bg = 3
cl = 0.90
mu_step = 0.001
print(fc.get_upper_limit(n, bg, cl, mu_step))

Output: 4.424
FC Table 4: 4.42
"""


import numpy as np
from scipy.stats import poisson
from scipy.special import erfinv


def get_lnR(n: int, mu: float, bg: float) -> float:
    """Computes the log of the FC ratio.

    Args:
        n: observed counts
        mu: expected signal counts
        bg: expected background counts

    Returns:
        lnR: log of L(n|mu+bg) / L(n|mu_best + bg)
    """
    if n == 0 and bg == 0: return -mu
    if mu + bg == 0:
        if n == 0: return 0
        return -np.inf
    mu_best = np.maximum(n-bg, 0)
    return n*(np.log(mu+bg) - np.log(mu_best+bg)) + mu_best - mu


def get_R(n, mu, bg):
    """Exponentiation of get_lnR"""
    return np.exp(get_lnR(n, mu, bg))


def get_equalR_mu(nLo: int, nHi: int, bg: float) -> float:
    """Finds the mu that solves R(nLo, mu, bg) = R(nHi, mu, bg).

    This seemingly unnecessary operation offers a huge speed up for computing FC
    limits. It can be computed analytically. The arguments nLo and nHi form sort
    of a "test" coverage region, and one searches for the mu returned by this
    function that has the "right" likelihood ratio value.

    Args:
        nLo: lower limit of "test" coverage region
        nHi: upper limit of "test" coverage region
        bg: expected background counts

    Returns:
        mu, the expected signal counts that solves
        R(nLo, mu, bg) = R(nHi, mu, bg)
    """
    if nLo >= nHi or nHi < bg: return 0
    if nLo == 0: return nHi / np.exp(1 - bg/nHi) - bg
    mu1 = np.maximum(nLo-bg, 0) + bg
    mu = np.exp((nHi*np.log(nHi)-nHi - nLo*np.log(mu1)+mu1)/(nHi-nLo)) - bg
    return mu


def get_edges(mu: float, bg: float, cl: float = 0.90) -> (int, int, float):
    """Finds the limits of the coverage patch (at cl) for mu+bg.

    Args:
        mu: expected signal counts
        bg: expected background counts
        cl: confidence level for the coverage patch, 90% by defaul

    Returns:
        The tuple (nLo, nHi, cov) containing the coverage and the lower and
        upper observed counts for the coverage patch of mu: solves Σ_nLo^nHi
        P(n|mu+bg) >= cl
    """
    if mu == 0: 
        nHi = poisson.ppf(cl, bg)
        return 0, nHi, poisson.cdf(nHi, bg)
    nLo = nHi = int(mu+bg)
    pLo = pHi = coverage = poisson.pmf(nLo, mu+bg)
    while coverage < cl:
        if nLo > 0 and get_lnR(nLo-1, mu, bg) > get_lnR(nHi+1, mu, bg):
            pLo *= nLo/(mu+bg)
            nLo -= 1
            coverage += pLo
        else:
            nHi += 1
            pHi *= (mu+bg)/nHi
            coverage += pHi
    return nLo, nHi, coverage


def get_lnRc(mu: float, bg: float, cl: float = 0.90) -> float:
    """Computes ln(Rc) for this mu.

    The critical ratio Rc is the minimum value of the likelihood ratio for n
    within the confidence interval for the given mu. The sum of P(n|mu+bg) for n
    with R >= Rc is as-close-as-possible but not lower than cl.

    Args:
        mu: expected signal counts
        bg: expected background counts
        cl: confidence level for the interval

    Returns:
        lnRc, the log of the critical ratio Rc
    """
    nLo, nHi, _ = get_edges(mu, bg, cl)
    return np.minimum(get_lnR(nLo, mu, bg), get_lnR(nHi, mu, bg))


def get_Rc(mu: float, bg: float, cl: float = 0.90) -> float:
    """Exponentiation of get_lnRc"""
    return np.exp(get_lnRc(mu, bg, cl))


def get_upper_limit_brute_force(n: int,
                                bg: float,
                                cl: float = 0.90,
                                mu_step: float = 0.001) -> float:
    """Gets FC upper limits, brute force method.

    Saved for posterity. This is a slow but straightforward computation of the
    Feldman-Cousins upper limit at confidence level cl when n counts are
    observed and bg background counts were expected.

    Args:
        n: observed counts
        bg: expected background counts
        cl: confidence level for the upper limit
        mu_step: precision for computing the upper limit

    Returns:
        mu, the upper limit on the signal strength, accurate to ±mu_step and
        rounded to the nearest interval of mu_step.
    """
    cl_sigma = erfinv(cl)*np.sqrt(2)
    mu = n + cl_sigma*np.sqrt(n)
    nLo, _, _ = get_edges(mu, bg, cl)
    while nLo > n and mu > 0:
        mu -= mu_step
        if mu < 0: mu = 0
        nLo, _, _ = get_edges(mu, bg, cl)
    while nLo <= n:
        mu += mu_step
        nLo, _, _ = get_edges(mu, bg, cl)
    return np.round(mu/mu_step)*mu_step


def get_upper_limit(n: int,
                    bg: float,
                    cl: float = 0.90,
                    mu_step: float = 0.001) -> float:
    """Gets FC upper limits, fast method.

    This is a fast but tricky computation of the Feldman-Cousins upper limit at
    confidence level cl when n counts are observed and bg background counts were
    expected.

    Args:
        n: observed counts
        bg: expected background counts
        cl: confidence level for the upper limit
        mu_step: precision for computing the upper limit

    Returns:
        mu, the upper limit on the signal strength, accurate to ±mu_step and
        rounded to the nearest interval of mu_step.
    """
    # FC is based on the ordering parameter R = P(n|mu+b) / P(n|mu_best+b)
    # where mu_best = max(0, n-b). One way to understand the confidence
    # intervals is to think of each mu as having a critical value R_c(mu|b) for
    # the ordering parameter, so that if n has R(n|mu,b) >= R_c(mu|b) then it is
    # inside the interval for that mu. When an experiment is done with known b
    # and gets result n, one then just needs to find the minimal / maximal mu
    # with R(mu|n,b) >= R_c(mu|b): these are the lower and upper FC limits.
    # However, computing R_c(mu|b) is time intensive, as it requires integrating
    # the Poisson distribution with variable limits.
    #
    # Consider the family of curves R(mu|n,b) as a function of mu:
    # R(mu|n,b) = (mu+b)^n exp(-(mu+b)) * const
    # These curves are monotonically decreasing for n < b, and for n > b rise to
    # a maximum at mu+b = n before falling monotonically beyond that.  The
    # R_c(mu|b) have to occur somewhere along these curves. R_c hops from curve
    # to curve for different n to achieve the coverage closest to CL without
    # undercovering. So the -last- R(mu|n,b) that is >= R_c(mu|b) (giving mu_UL)
    # will either be at or just above a point where R(mu|n,b) = R(mu|n',b) for
    # some n' > n. Moreover, these points require that R(mu|n,b) be falling and
    # R(mu|n',b) be rising, and the latter can only occur for n > b. A similar
    # argument can be made for lower limits.
    #
    # So: the upper limits can be more efficiently found by first stepping up in
    # n' from max(n+1, ceil(b)), and for each computing the mu' such that
    # R(mu|n,b) = R(mu|n',b), which is a one-step analytic calculation. One then
    # computes R_c(mu'), and stops stepping up in n' when R(mu'|n,b) < R_c(mu').
    # Having found a candidate for mu_UL, one can peek forward in mu to see if
    # higher mu have lower R_c. If so, a binary search can be performed between
    # there and the next higher mu' to find mu_UL.
    nprime = np.maximum(n+1, np.ceil(bg)) - 1
    lnR = lnRc = 0
    while lnR >= lnRc:
        nprime += 1
        mu = get_equalR_mu(n, nprime, bg)
        lnR = get_lnR(n, mu, bg)
        lnRc = get_lnRc(mu, bg, cl)
    nprime -= 1
    muLo = get_equalR_mu(n, nprime, bg) + mu_step
    lnR = get_lnR(n, muLo, bg)
    lnRc = get_lnRc(muLo, bg, cl)
    if lnR < lnRc: return muLo - mu_step
    # if we get here, do a binary search between mu(n') and mu(n'+1)
    muHi = get_equalR_mu(n, nprime+1, bg)
    while muHi-muLo > mu_step:
        mu = (muLo + muHi)/2
        lnR = get_lnR(n, mu, bg)
        lnRc = get_lnRc(mu, bg, cl)
        if lnR < lnRc: muHi = mu
        else: muLo = mu
    return np.round(muLo/mu_step)*mu_step


def get_lower_limit(n: int,
                    bg: float,
                    cl: float = 0.90,
                    mu_step: float = 0.001) -> float:
    """Gets FC lower limits, fast method.

    This is a fast but tricky computation of the Feldman-Cousins lower limit at
    confidence level cl when n counts are observed and bg background counts were
    expected. See code comments in get_upper_limit for the algorithm.

    Args:
        n: observed counts
        bg: expected background counts
        cl: confidence level for the lower limit
        mu_step: precision for computing the lower limit

    Returns:
        mu, the lower limit on the signal strength, accurate to ±mu_step and
        rounded to the nearest interval of mu_step.
    """
    # First check n = 0 or mu = 0, to see if we can quickly exit.
    if n == 0: return 0

    mu = 0
    lnR = get_lnR(n, mu, bg)
    lnRc = get_lnRc(mu, bg, cl)
    if lnR >= lnRc: return 0

    # Now start from n and walk backwards. We should not get to zero!
    nprime = n
    lnR = lnRc = 0
    while lnR >= lnRc and nprime > 0:
        nprime -= 1
        mu = get_equalR_mu(nprime, n, bg)
        lnR = get_lnR(n, mu, bg)
        lnRc = get_lnRc(mu, bg, cl)
    if nprime > 0 or lnR < lnRc: nprime += 1
    muHi = get_equalR_mu(nprime, n, bg) - mu_step
    lnR = get_lnR(n, muHi, bg)
    lnRc = get_lnRc(muHi, bg, cl)
    if lnR < lnRc: return muHi + mu_step
    # if we get here, do a binary search between mu(n'-1) and mu(n')
    if nprime > 0: muLo = get_equalR_mu(nprime-1, n, bg)
    else: muLo = 0
    while muHi-muLo > mu_step:
        mu = (muLo + muHi)/2
        lnR = get_lnR(n, mu, bg)
        lnRc = get_lnRc(mu, bg, cl)
        if lnR < lnRc: muLo = mu
        else: muHi = mu
    return np.round(muLo/mu_step)*mu_step


def median_limit_sensitivity(bg: float,
                             cl: float = 0.90,
                             mu_step: float = 0.001) -> float:
    """Computes the FC median limit sensitivity.

    The median limit sensitivity is the upper limit an experiment with
    background level bg would result in if the observed number of counts were
    equal to the median of the expected count distribution when the signal is
    too small to be measured (mu << bg)

    Args:
        bg: expected background counts
        cl: confidence level for the upper limit
        mu_step: precision for computing the lower limit

    Returns:
        mu, the median limit sensitivity, accurate to ±mu_step and rounded to
        the nearest interval of mu_step.
    """
    return get_upper_limit(poisson.median(bg), bg, cl, mu_step)


def mean_limit_sensitivity(bg: float,
                           cl: float = 0.90,
                           mu_step: float = 0.001) -> float:
    """Computes the FC mean limit sensitivity.

    The mean limit sensitivity is the average of the upper limits an ensemeble
    of identical experiment with background level bg would result in if the
    signal is too small to be measured (mu << bg) so that the observed counts
    would be Poisson distributed with mean = bg.

    Args:
        bg: expected background counts
        cl: confidence level for the upper limit
        mu_step: precision for computing the lower limit

    Returns:
        mu, the mean limit sensitivity, accurate to ±mu_step and rounded to the
        nearest interval of mu_step.
    """
    get_upper_limit_vec = np.vectorize(get_upper_limit)
    def f(n):
        return get_upper_limit_vec(n, bg, cl, mu_step)
    return poisson.expect(f, [bg])


def median_discovery_sensitivity(bg: float,
                                 cl: float = 0.9973,
                                 mu_step: float = 0.001) -> float:
    """Computes the discovery sensitivity.

    The discovery sensitivity is the smallest signal strength mu such that an
    ensemeble of identical experiment with background level bg would report a
    discovery (mu > 0) at cl at least 50% of the time.

    This concept does not make use of the FC ordering parameter. However, it can
    be quickly calculated using the FC construction, because the confidence band
    for mu=0 corresponds to the background-only hypothesis. So the first mu that
    gives a poisson median n_med that is not inide this interval is the
    discovery sensitivity.

    Args:
        bg: expected background counts
        cl: confidence level for the upper limit
        mu_step: precision for computing the lower limit

    Returns:
        mu, the discovery sensitivity, accurate to ±mu_step and rounded to the
        nearest interval of mu_step.
    """
    _, nHi, _ = get_edges(0, bg, cl)
    n_crit = nHi + 1
    mu = n_crit
    dmu = 1
    # now do a binary search for mu within n_crit ± dmu
    while dmu > mu_step:
        if poisson.median(mu) < n_crit:
            mu += dmu
        else:
            mu -= dmu
        if mu < 0: mu = 0
        dmu /= 2
    return np.round((mu - bg)/mu_step)*mu_step



"""
A package for setting frequenstist limits from unbinned data
"""

from freqfit.dataset import Dataset
from freqfit.experiment import Experiment
from freqfit.limit import SetLimit
from freqfit.superset import Superset
from freqfit.toy import Toy

__all__ = [
    "Dataset",
    "Experiment",
    "SetLimit",
    "Superset",
    "Toy",
    "PlotLimit" "__version__",
]



import numba as nb
import numpy as np

from freqfit.utils import inspectparameters

nb_kwd = {
    "nopython": True,
    "parallel": False,
    "nogil": True,
    "cache": True,
    "fastmath": True,
    "inline": "always",
}

SEED = 42  # set the default random seed

@nb.jit(**nb_kwd)
def nb_pdf(
    Es: np.array,
    S: float,
    B: float,
) -> np.array:

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        if Es[i] == 1:
            y[i] = 1.0
        else:
            y[0] = 0.0
                
    return y

@nb.jit(**nb_kwd)
def nb_density(
    Es: np.array,
    S: float,
    B: float,
) -> np.array:

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        if Es[i] == 1:
            y[i] = S + B
        else:
            y[0] = 0.0
                
    return S + B, y

@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_extendedrvs(
    S: float,
    B: float,
    seed: int = SEED,
) -> np.array:

    np.random.seed(seed)

    n_sig = np.random.poisson(S + B)
    
    Es = np.ones(n_sig)

    return Es, (n_sig, 0)

class onebin_poisson_gen:
    def __init__(self):
        self.parameters = inspectparameters(self.density)
        pass

    def pdf(
        self,
        Es: np.array,
        S: float,
        B: float,
    ) -> np.array:
        return nb_pdf(Es, S, B)

    # for iminuit ExtendedUnbinnedNLL
    def density(
        self,
        Es: np.array,
        S: float,
        B: float,
    ) -> np.array:
        return nb_density(Es, S, B)

    def extendedrvs(
        self,
        S: float,
        B: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_extendedrvs(S, B, seed=seed)

onebin_poisson = onebin_poisson_gen()



"""
This is the script that actually runs toys for a given value of S
"""
# import os module
import os

# turn off file locking
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

# import python modules
import argparse

from freqfit.limit import SetLimit
from freqfit.utils import load_config

# Setup the argument parser
__pars__ = argparse.ArgumentParser()

# include the arguments
__pars__.add_argument("-i", type=str, default="", help="path to the input yaml file")
__pars__.add_argument(
    "-s", type=str, default="0.055", help="signal rate to evaluate toys at"
)
__pars__.add_argument("-j", type=int, default=1, help="job id in the job array")
__pars__.add_argument(
    "-o",
    type=str,
    default="/home/sjborden/freqfit/gerda_toys",
    help="output path to a folder to dump all the files in",
)
__pars__.add_argument("-n", type=int, default=1, help="number of toys to run")

# parse the arguments
__args__ = __pars__.parse_args()

# interpret the arguments
# proc_count = __args__.c if __args__.c < mp.cpu_count() else mp.cpu_count()
proc_input = __args__.i
proc_s = __args__.s
proc_output = __args__.o
proc_job_id = __args__.j
proc_num_toy = __args__.n


if __name__ == "__main__":
    print("Starting program!", flush=True)  # noqa:T201

    p = SetLimit(
        load_config(proc_input),
        proc_job_id,
        proc_num_toy,
        proc_output,
        name="l200_toys",
    )

    p.set_var_to_profile("global_S")

    print(f"profiling {float(proc_s)}", flush=True)  # noqa:T201

    p.run_and_save_toys(float(proc_s))

    print("exiting program!", flush=True)  # noqa:T201
